function app = llrdisc(model, cfg)
% LLR approximator with discrete actions
% Used for easy handling of the output of llrfittedqi

CFG.type = 'llrdisc';
CFG.disc = 1;           % always discrete
CFG.U = [];             % cell array of grids
% samples
CFG.Xs = [];            % state samples
CFG.Is = [];            % indices of samples for each discrete action (M-length cell array)
CFG.Us = [];            % alternative to the above: give actual samples
CFG.Qs = [];            % default Q-function samples (can be overridden)
% LLR
CFG.knnmethod = [];     % like llrfittedqi, 'kdtree' or 'jit'
CFG.knn = [];           % how many nearest neighbors
CFG.S = [];             % distance scaling; MUST be specified explicitly (does not support 'auto' etc.)
CFG.wfun = [];          % weight fun
CFG.kdtrees = [];       % pointers to existing KD trees for KNN computation (otherwise they are created)
CFG.linsolvemethod = [];% 'pinv' (consistent with llrfittedqi), or 'mldivide' - more efficient
CFG.extrap = [];        % how to handle extrapolation

global KDTREE_POINTERS  % global memory of tree pointers, for potential deletion later on

app = parseconfig(cfg, CFG);

% some error checks
if isempty(app.U),
    error('LLRDISC requires predefined discrete actions in field U');
elseif isempty(app.Xs) || (isempty(app.Is) && isempty(app.Us)),
    error('LLRDISC requires state samples Xs and action indices Is or samples Us)');    
end;
if ~any(strcmp(app.linsolvemethod, {'mldivide', 'pinv'})),
    error('Linear solve method [%s] unkown', app.method);
end;

app.Uflat = flat(app.U);
app.maxx = model.maxx; app.maxu = model.maxu;
% number of action variables and discrete actions
app.dimx = model.p; app.dimu = model.q;
app.M = size(app.Uflat, 2);

% If needed, identify portion of sample data corresponding to each discrete action
if isempty(app.Is),
    app.Is = cell(app.M, 1);
    for j = 1:app.M, app.Is{j} = findflat(app.Uflat(:, j), app.Us); end;
end;

% if needed, create trees
if strcmp(cfg.knnmethod, 'kdtree') && isempty(app.kdtrees),
    app.kdtrees = nan(app.M, 1);
    for j = 1:app.M, app.kdtrees(j) = kdtree_build(app.Xs(:, app.Is{j})'); end;
    KDTREE_POINTERS(end+1:end+app.M) = app.kdtrees;
end;

% there is no "activation degree" function because the approximator is not
% linear
app.phi = []; 
app.q = @llrdisc_q; app.h = @llrdisc_h; 
% plot functions
app.plotv = @llrdisc_plotv; app.ploth = @llrdisc_ploth;
app.vectorized = 1;    % signal approximator supports vectorized q and h calls

end     % constructor

% ===== Core functions

% Q-value in x, u (theta is currently ignored since none of the supported reg use it)
% supports vectorized x and u (the # of points should obviously match)
function q = llrdisc_q(app, Qs, x, u)
if isempty(Qs), Qs = app.Qs; end;
Np = size(x, 2); q = zeros(1, Np);
% Xs = app.Xs; Uflat = app.Uflat;
% knn = app.knn; methodflag = app.knnmethod(1);
% S = app.S; dimx = app.dimx;
for i = 1:Np,
    % find discrete action index
    j = findflat(u(:, i), app.Uflat, 1, 'first');
    % kNN search
    if app.knnmethod(1) == 'k',     % kdtree CPP implementation
        [ind dist] = kdtree_k_nearest_neighbors(app.kdtrees(j), x(:, i), app.knn);
        ind(end:-1:1) = app.Is{j}(ind); dist(end:-1:1) = dist; 
    else                            % Matlab JIT assumed
        try
            [ind dist] = knnsearchx(x(:, i)', app.Xs(:, app.Is{j})', app.knn, app.S);
            ind = app.Is{j}(ind);
        catch
            disp(x(:, i)')
            sizeXs = size(app.Xs(:, app.Is{j})')
            knn = app.knn
            S = app.S
        end;
    end;
    qllr = Qs(ind);
    w = llrweights(app.wfun, dist, dist(end))';
    % for each kNN: subtract query sample and multiply by weight
    xllr = zeros(app.knn, app.dimx+1);
    for ik = 1:app.knn,
        xllr(ik, 1:app.dimx) = w(ik) .* (app.Xs(:, ind(ik)) - x(:, i));
    end;
    % bias column = weights
    xllr(:, end) = w;
    % solve linear system
    if app.linsolvemethod(1) == 'p',        % pinv
        betallr = pinv(xllr) * (w .* qllr);
    else                                    % mldivide assumed
        betallr = xllr \ (w .* qllr);
    end;
    % since query sample is zero-ed, regression output = bias term
    if app.extrap(1) == 's',                % extrapolation handling: saturate each sample
        q(i) = max(min(qllr), min(betallr(end), max(qllr))); 
    else        % don't saturate
        q(i) = betallr(end);
    end;
end;
end

% Greedy policy in x
% Supports vectorized x. Also returns the Q-values of the optimal actions
% could be optimized for looping over j
function [u, qmax] = llrdisc_h(app, Qs, x)
if isempty(Qs), Qs = app.Qs; end;
Np = size(x, 2); 
if Np == 1,
    q = llrdisc_q(app, Qs, repmat(x, 1, app.M), app.Uflat);
    [qmax imax] = max(q);
    u = app.Uflat(:, imax); 
    for i = 1:Np, u(:, i) = app.Uflat(:, imax(i)); end;
else
    xu = flatcrossprod(x, app.Uflat);
    q = llrdisc_q(app, Qs, xu(1:app.dimx, :), xu(app.dimx+1:end, :));
    % note because of how flatcrossprod works, we need to invert dimensions in the result:
    q = reshape(q, app.M, Np)';
    [qmax imax] = max(q, [], 2);
    u = zeros(app.dimu, Np); 
    for i = 1:Np, u(:, i) = app.Uflat(:, imax(i)); end;
end;
end

% ----- Plot functions
% Plot V - forward call to generic plot fun
function llrdisc_plotv(app, Qs, varargin)
approx_plotv(app, Qs, [-app.maxx app.maxx], [-app.maxu app.maxu], varargin{:});
end
% Plot h - forward call to generic plot fun
function llrdisc_ploth(app, Qs, varargin)
approx_ploth(app, Qs, [-app.maxx app.maxx], [-app.maxu app.maxu], varargin{:});
end
