function varargout = montecarlo(cfg)
% Q-learning
%   VARARGOUT = montecarlo(CFG)
% Inputs:
%   CFG             - structure with fields as commented in the code below
%           can also be given as a string, see he str2cfg
% Outputs:
%   Q               - in run mode. Computed Q-function.
%   HIST, FIGH      - in replay mode. HIST is the replay history.
%           FIGH contains figure handles if figures were created 
%           and not closed; and is an empty matrix if all the figures were closed.

if nargin < 1, cfg = struct(); end;

% -----------------------------------------------
% Process configuration structure

% default config
% script config
CFG.model_params = {};              % extra parameters for problem calling in 'model' mode
CFG.run = 0;                        % run learning
CFG.replay = 0;                     % replay learned policy
CFG.problem = '';                   % what problem to solve
CFG.datadir = '';                   % data dir
CFG.datafile = '';                  % save data to file
CFG.gamma = 0.98;                   % discount factor
CFG.explor = 1;                     % exploration rate
CFG.explordecay = .98;              % exploration decay (per trial)
CFG.maxtrials = 100;                % max # trials
CFG.maxsteps = 1000;                % max # steps in a trial
CFG.peval_trials = 1;               % how many trials of policy evaluation before policy improvement
CFG.visit = 'first';                % 'first', or 'every'
CFG.explortype = 'step';            % 'es'=exploring starts, 'step'=at every step
CFG.resetreturns = 0;               % whether returns should be reset on policy improvement
CFG.h0 = [];                        % initial policy
CFG.reset = 'rand';                 % trial reset
CFG.randseed = [];                  % set the random seed (for repeatable results)
% replay config
CFG.x0 = [];                        % initial state for replay (otherwise the problem default or zeros)
CFG.tend = 30;                      % end time for replay

CFG.plottarget = 'screen';          % 'screen', '', 'latex', or 'beamer'. If 'screen' figures will not be closed
CFG.savedir = '';
CFG.savefig = '';
% display config
CFG.verb = 5;                       % verbosity: the higher, the more detailed the messages displayed
CFG.visualize = 0;                  % visualization level (0 = none, 1 = trial-level, 2 = step-level)
CFG.viscfg = struct;                % visualization config options
CFG.trialdisp = 1;

% List of fields that define the problem and therefore may NOT be overwritten on load
KEEPFIELDS = {'problem', 'gamma'};

% Install function defaults for everything else
cfg = parseconfig(cfg, CFG);

cfg.init = cfg.run || ~exist([cfg.datafile '.mat'], 'file');

if ~cfg.init,        % load data file, making sure that cfg is not overwritten
    cfg1 = cfg; kf = KEEPFIELDS;
    load(cfg.datafile);
    % Overwrite problem-defining fields from loaded config, keep the rest as in the (current) cfg1;
    % the result becomes the new config
    cfg = copyfields(cfg, cfg1, kf);
    KEEPFIELDS = kf; clear cfg1 kf;
    dispx(['Data loaded from [' cfg.datafile '].'], cfg.verb, 1);
end;

% Echo config
dispx('Q-learning called with the following configuration:', cfg.verb, 1);
dispx(cfg, cfg.verb, 1);

% get environment (Matlab, hardware) info
cfg.envinfo = getenvx;

% -----------------------------------------------
% Create model, find dimensionality data
if cfg.init,
    model = feval(cfg.problem, 'model', cfg.model_params{:});
    DIMS.dimx = zeros(model.p, 1);
    DIMS.dimu = zeros(model.q, 1);
    for i = 1:model.p, DIMS.dimx(i) = length(model.X{i}); end;
    for i = 1:model.q, DIMS.dimu(i) = length(model.U{i}); end;
    DIMS.N = prod(DIMS.dimx);
    DIMS.M = prod(DIMS.dimu);
    Xflat = flat(model.X); Uflat = flat(model.U);
    p = model.p; q = model.q;
end;

% -----------------------------------------------
% Q-iteration
if cfg.run,
    
    cfg.maxiter = cfg.maxtrials/cfg.peval_trials;   % total number of policy iterations
    
    % init Q-function, elig trace, histories etc.
    X = NaN(p, cfg.maxsteps+1, cfg.maxtrials);
    U = NaN(q, cfg.maxsteps+1, cfg.maxtrials);
    R = NaN(1, cfg.maxsteps+1, cfg.maxtrials);
    RET = zeros(DIMS.N, DIMS.M);        % cumulative returns, divided by number => Q
    RETn = zeros(DIMS.N, DIMS.M);       % number of visits
    % history of policies
    hh = cell(cfg.maxiter+1, 1);
    % a linear history of Q functions: for each policy
    Qh = cell(cfg.maxtrials+1, 1); 

    deltah = NaN(cfg.maxiter+1, 1);
    peval_deltah = NaN(cfg.maxiter, cfg.peval_trials+1);

    % init Q
    Q = zeros(DIMS.N, DIMS.M);
    Qh{1} = Q;
    
    % initialize policy
    if isempty(cfg.h0), h = repmat(Uflat(:, 1), 1, DIMS.N); % always first action
    else                h = cfg.h0;
    end;
    hh{1} = h;

    trial = 1;
    timestat.run = 0;
    if cfg.explortype(1) == 's', explor = cfg.explor; end;
    if cfg.explortype(1) == 'e' || ~strcmp(cfg.reset, 'rand'),
        dispx('Warning! Exploring starts -- seting reset to rand', cfg.verb, 0);
        cfg.reset = 'rand';
    end;
    
    if ~isempty(cfg.randseed), rand('twister', cfg.randseed); end;
    
    % init visualization config if needed
    if cfg.visualize,
        vcfg = cfg.viscfg;
        vcfg.gview = [];
    end;

    % -----------------------------------------------
    % Perform MC
    dispx('Performing Monte Carlo learning...', cfg.verb, 0);

    t = cputime;
    ell = 1; peval_trial = 1;
    explor = cfg.explor;
    while trial <= cfg.maxtrials,       % main loop

        % initialize trial
        k = 1; terminal = 0;
        % initial state
        X(:, 1, trial) = feval(cfg.problem, 'reset', model, cfg.reset);

        timestat.run = timestat.run + (cputime - t);
        % visualize initial state
        if cfg.visualize >= 2,
            vcfg.mcstep = 1;
            vcfg.mctrial = 0;
            vcfg.k = k - 1;
            vcfg.trial = trial;
            [figh vcfg.gview] = feval(model.visualizefun, vcfg);
        end;
        t = cputime;
            
        while k < cfg.maxsteps && ~terminal,
            % if exploring starts, we only explore at the first step
            if cfg.explortype(1) == 'e' && k > 1, exploring = 0;
            else exploring = rand < explor;
            end;
            % choose action
            if exploring,   
                U(:, k, trial) = Uflat(:, unidrnd(DIMS.M));
            else
                xind = findflat(X(:, 1, trial), Xflat, 1, 'first');
                U(:, k, trial) = h(xind);
            end;
            
            % apply the action
            [X(:, k+1, trial), R(1, k+1, trial), terminal] = ...
                feval(model.fun, model, X(:, k, trial), U(:, k, trial));
            
            % update stats
            timestat.run = timestat.run + (cputime - t);
            
            % visualization
            if cfg.visualize >= 2,
                vcfg.mcstep = 1;
                vcfg.mctrial = 0;
                vcfg.k = k;
                vcfg.trial = trial;
                vcfg.ell = ell;
                vcfg.peval_trial = peval_trial;
                vcfg.policyimproved = 0;
                [figh vcfg.gview] = feval(model.visualizefun, vcfg);
            end;
            
            t = cputime; 
            k = k + 1;
        end; %end of trial
        if ~terminal, dispx('Warning! Trial ended without terminal state', cfg.verb, 0); end;
        
        % calculate the returns for each state in the trajectory
        K = k - 1;
        discounting = cfg.gamma .^ (0:K-1);
        returns = R(1, 2:K+1, trial) .* discounting;
        % [gamma^0*R_0 gamma^1*R_1 ... gamma^K*R_K gamma^K+1*Rinf] (need to reverse to do the cumulative sum)
        returns = cumsum(returns(end:-1:1));
        % divide R_k by gamma^k for each k,to find out return starting from k
        returns = returns(end:-1:1) ./ discounting;        

        % update cumulative returns and counters
        skip = [];
        for k = 1:K,
            if any(k == skip), continue; end;
            x = X(:, k, trial); xind = findflat(x, Xflat, 1, 'first');
            u = U(:, k, trial); uind = findflat(u, Uflat, 1, 'first');
            RET(xind, uind) = RET(xind, uind) + returns(k);
            RETn(xind, uind) = RETn(xind, uind) + 1;
            if cfg.visit(1) == 'f',
                % skip all other visits to the same x-u pair
                skip = [skip findflat([x; u], [X(:, k:K, trial); U(:, k:K, trial)])];
            end;
        end;
        
        %update Q based on average returns, and store on history
        visited = RETn > 0;
        Q(visited) = RET(visited) ./ RETn(visited);
        Q(~visited) = 0;
        Qh{trial+1} = Q;
        
        % compute difference, but only if we evaluate the same policy OR we update every single
        % trial
        if peval_trial > 1 || cfg.peval_trials == 1,
            peval_deltah(ell, peval_trial+1) = max(max(abs(Q - Qh{trial})));
        end;
        
        % policy improvement
        improvepolicy = ~mod(trial, cfg.peval_trials);
        if improvepolicy,
            [Qmax hind] = max(Q, [], 2);
            h = Uflat(:, hind);
            hh{ell+1} = h;
            deltah(ell+1) = sum(~all(h - hh{ell}));
            % reset returns if configured
            if cfg.resetreturns, RET(:) = 0; RETn(:) = 0; end;
        end;
        
        % update stats
        timestat.run = timestat.run + (cputime - t);
        
        % console feedback of algorithm progress
        if ~mod(trial, cfg.trialdisp),
            if improvepolicy, pistatus = ' [done]'; else pistatus = ''; end;
            dispx(sprintf('Trial #%d done - pevaltrial #%d for piter #%d%s', ...
                trial, peval_trial, ell, pistatus), cfg.verb, 2);
        end;
        % visualization
        if cfg.visualize >= 1,
            vcfg.mctrial = 1;
            vcfg.mcstep = 0;
            vcfg.trial = trial;
            vcfg.ell = ell;
            vcfg.peval_trial = peval_trial;
            vcfg.policyimproved = improvepolicy;
            [figh vcfg.gview] = feval(model.visualizefun, vcfg);
        end;
        
        % start counting time again, increment trial counters
        t = cputime;
        if improvepolicy, ell = ell + 1; peval_trial = 0; end;
        trial = trial + 1;
        peval_trial = peval_trial + 1;
        % anneal exploration
        explor = explor * cfg.explordecay;
    end;        % while not converged and allowed more iterations

    dispx(sprintf('Algorithm stopped, %d trials run.', cfg.maxtrials), cfg.verb, 0);
    
    % finalize visualizer
    if cfg.visualize,
        vcfg.mctrial = 0;
        vcfg.mcstep = 0;
        vcfg.finalize = 1;
        [figh vcfg.gview] = feval(model.visualizefun, vcfg);
    end;
    
    % output Q-function
    varargout = {Q};
end;

% -----------------------------------------------
% Backup data

if cfg.run && ~isempty(cfg.datafile),
    % save into the same directory as the problem
    if isempty(cfg.datadir),
        datadir = fileparts(which(cfg.problem));
    else
        datadir = cfg.datadir;
        if (datadir(end) == '\'), datadir = datadir(1:end-1); end;
    end;
    cfg.datafile = [datadir '\' cfg.datafile];
    save(cfg.datafile);
    dispx(['Data was saved to [' cfg.datafile '].'], cfg.verb, 1);
end;


% END qi() RETURNING varargout =================================================================
