% This script demonstrates the use of several online, optimistic planning algorithms
% using the inverted pendulum swingup problem

% The file is best uses in cell mode, place the cursor in a cell (i.e. range between two "%%" signs)
% and press CTRL+ENTER

% cleanup code
close all; clear all; clc;
delete ip_opd.mat ip_opss.mat ip_soop.mat

%% OPD - Optimistic Planning for Deterministic Systems
% experiment configuration
cfg = struct;

% first we need to configure which problem we apply the algorithm to
cfg.problem = 'ipsetup_problem';
% some configuration for the problem, see ipsetup_problem and the help for the 'model' mode
% first argument of the model mode is the reward: "normalized LQR"
%       -- quadratic in states with weights Qrew and in actions with weights Rrew, 
%       but scaled and translated into [0,1]
% second argument is the physical params, we don't change anything there
% third argument is the numerical integration, we set a sampling time of Ts=.05, somewhat larger
% than usual to make it easier for planning to get to longer horizons
cfg.model_params = {'rewtype=nlqr Qrew=[1,0] Rrew=0.3', [], 'Ts=.05'}; 

% discount factor
cfg.gamma = 0.95;

% next come the settings for the planning at each step
% note the function called in the end is actually genmpc, a wrapper that works for any planning
% algo, it takes the algo and its params as an argument and then takes care of applying it in closed
% loop
cfg.planalg = 'opd';            % planning algo: optimistic planning for deterministic systems
% plan algo parameters
cfg.planparams.n = 50;          % budget of expansions at each step
cfg.planparams.U = [-3 0 3];    % action space: small enough amplitude to require a swing
cfg.planparams.verb = 0;        % verbosity during planning process: we don't want to see anything
cfg.reset = [-pi, 0]';          % reset pointing down
cfg.trials = 1;                 % one trial is enough
cfg.maxsteps = 60;              % 60 steps = 3 seconds should be enough for a swingup

% finally some other misc config arguments
% output config
cfg.datadir = pwd;              % output data file to current dir
cfg.datafile = 'ip_opd';        % save data file with this name
cfg.verb = 0;                   % there's also a verbosity setting at the control level
% we can also use a visualizer to see a graphical representation of the system as it's being
% controlled
% set visualize to 0 to disable
cfg.visualize = 3;             % this means: only visualize system steps
cfg.viscfg.pause = 0.001;      % pause a bit each step
cfg.viscfg.figsize = [450 600];% figure size for visualizer

% run the algorithm
cfg.run = 1;
genmpc(cfg);              


% Now that we have the results, we can inspect them
% look at the controlled trajectory
genmpc('plottrial datafile=ip_opd');


%% OPSS - Optimistic Planning for MDPs (or for 'Sparsely Stochastic' Systems)
% we don't comment things that are the same as before
cfg = struct;
% problem config
cfg.problem = 'ipsetup_problem';
% note now we use a stochastic variant of the inverted pendulum, where the actuator applies the
% intended action with proba 0.7, and an action with 0.8 the intended magnitude with proba 0.3
cfg.model_params = {'rewtype=nlqr Qrew=[1,0] Rrew=0.3', 'stoch_prob=[.7,.3] stoch_gain=[1,.8]', 'Ts=.05'}; 
cfg.gamma = 0.95;
cfg.planalg = 'opss';           
% plan algo config
cfg.planparams.n = 200;         % budget of expansions, each expansions twice as expensive as in OPD
cfg.planparams.U = [-3 0 3];    
cfg.planparams.L = 2;           % how many stochastic outcomes are possible at each transition
cfg.planparams.verb = 0;        
cfg.reset = [-pi, 0]';          
cfg.trials = 1;                 
cfg.maxsteps = 60;              
% other fields
cfg.datadir = pwd;             
cfg.datafile = 'ip_opss';   
cfg.verb = 3;                   % verbosity 3: show state, action, and planning summary info at each step 
cfg.visualize = 3;             
cfg.viscfg.pause = 0.001;      
cfg.viscfg.figsize = [450 600];
% run
cfg.run = 1;
genmpc(cfg);              

% look at the controlled trajectory
genmpc('plottrial datafile=ip_opss');

%% Tree reuse
% note that both OPD and OPSS can reuse trees at subsequent steps, or build a b-value approximator, 
% we can use this to get better performance with smaller budgets

% run the above first to get the OPSS config
cfg.planparams.reusetree = 1;               % signal OPSS that the tree should be reused
cfg.planparams.reusetree_maxsize = 3000;    % beyond this size the tree is cleared from memory
cfg.planparams.n = 175;

cfg.run = 1;
genmpc(cfg);  

%% Memory reuse
% run the above first to get the OPSS config
cfg.planparams.reusetree = 0;               % can't reuse tree and memory at the same time
cfg.planparams.usemem = 1;                  % signal OPSS that it should reuse the memory
cfg.planparams.mem_maxsize = 1500;          % keep memory at this size by removing samples
cfg.planparams.lipschitz = 0.5;             % approximator assumes a Lipschitz b-value with this constant
cfg.planparams.n = 150;

cfg.run = 1;
genmpc(cfg);  

%% SOOP - SOO for Plannning, a continuous-action algorithm
cfg = struct;
% problem config
cfg.problem = 'ipsetup_problem';
cfg.model_params = {'rewtype=nlqr Qrew=[1,0] Rrew=0.3', [], 'Ts=.05'}; 
cfg.gamma = 0.95;
cfg.planalg = 'sooplp';             % the name is sooplp because the algo function also implements LP
% plan algo config
cfg.planparams.method = 'soop';     % since sooplp also implements LP, tell it which algo to use
cfg.planparams.n = 1000;            % budget for SOOP is given as a count of model calls 
cfg.planparams.dimselect = 'disc';  % selection method for which dimension to split: discounting-based
cfg.planparams.alpha = 0.8;         % discount factor for dimension selection
cfg.planparams.verb = 0;        
cfg.reset = [-pi, 0]';          
cfg.trials = 1;                 
cfg.maxsteps = 60;              
% other fields
cfg.datadir = pwd;             
cfg.datafile = 'ip_soop';   
cfg.verb = 0;                      
cfg.visualize = 3;             
cfg.viscfg.pause = 0.001;      
cfg.viscfg.figsize = [450 600];
% run
cfg.run = 1;
genmpc(cfg);              

% look at the controlled trajectory
genmpc('plottrial datafile=ip_soop');


%% Running planning in offline mode
% Note we can also run any planning algorithm in offline mode, for a grid of initial states
cfg = struct;
% problem config
cfg.problem = 'ipsetup_problem';
cfg.model_params = {'rewtype=nlqr Qrew=[1,0] Rrew=0.3', [], 'Ts=.05'}; 
cfg.gamma = 0.95;
cfg.planalg = 'sooplp';              % the name is sooplp because the algo function also implements
% offline mode config
cfg.online = 0;
% set of initial states: usually a grid, which can be given as separate vectors along each state
% dimension; a crossproduct of these vectors will then be used
cfg.X = {-pi:pi/3:pi,-4*pi:pi:4*pi};
% plan algo config
cfg.planparams.method = 'soop';     
cfg.planparams.n = 1000;            
cfg.planparams.dimselect = 'disc';
cfg.planparams.alpha = 0.8;
cfg.planparams.verb = 0;        
% other fields
cfg.datadir = pwd;             
cfg.datafile = 'ip_soop';   
cfg.verb = 3;                   % verbosity 3: show state, action, and planning summary info at each step 
cfg.visualize = 3;             
cfg.viscfg.pause = 0.001;      
cfg.viscfg.figsize = [450 600];
% run
cfg.run = 1;
genmpc(cfg);              

% we could then look at the actions for each state, using mode 'ploth' (from "plot policy")
genmpc('ploth datafile=ip_soop');

