function [ind, mu, indp] = mdegs_1(x, c, roll, n, p, tab)
%  Computes the membership degrees in a 1-dimensional array of triangular membership functions.
%   [IND, MU, INDP] = MDEGS_P(X, C, ROLL, N, P, TAB)
%
%  Computes membership degrees for a variable in a 1-dimensional array of
%  triangular membership functions. These fuzzy sets sets are normal and
%  the support of each fuzzy set is the interval defined by the centers of the two adjacent
%  fuzzy sets, such that the sum of membership degrees is always one. Such a collection of 1D
%  fuzzy sets is completely determined by the array of centers.
%
%  Parameters:
%   X           - the variable value
%   C           - the centers of the fuzzy sets. A 1-element cell array, with
%       C{1} a row vector containing the triangle centers.
%       The elements of the vector must be unique and sorted in ascending order.
%   ROLL        - rollover flag for the membership functions. Currently not used, reserved.
%   N           - the numbers of fuzzy sets, N = length(C{1})
%   P           - must be 1. Kept for compatibility with the mdegs_* functions
%   TAB         - unused. Kept for compatibility with the mdegs_* functions
%   
%  Returns:
%   IND         - the indices of the activated p-dimensional membership degrees. A vector
%       of length 2. The activated membership degrees are usually the nonzero membership
%       degrees. For ease of use and implementation, however, some zero membership degrees will
%       be regarded as activated to keep the length of ind constant.
%   MU          - the vector of activated membership degrees (some of these may be zero). The
%       membership degrees in MU correspond 1-to-1 with the indices in IND.
%   INDP        - identical to IND. Kept for compatibility with the mdegs_* functions

% Author:   Lucian Busoniu
% Version:  1.0
% Date:     2007-07-03

% find i such that x \in [c(i), c(i+1)); i=n if x = last center value
i = find(c{1} <= x, 1, 'last');
if i < n,   	% i, i+1 MFs are activated
    m  = (c{1}(i+1) - x) / (c{1}(i+1) - c{1}(i));
    mu = [m;  1-m];         % activation degrees
    ind = [i;  i+1];        % and indices
else            % singleton on the last MF
    mu = [  0;  1];         % activation degrees (first is a dummy)
    ind = [i-1;  i];        % and indices (first is a dummy)
end;

indp = ind;                

% END mdegs_n RETURNING ind, mu, indp ---------------------------

