function  varargout = malinear_problem(mode, varargin)
% Markov game with decoupled linear dynamics and homogeneous agents.
%   OUT = MALINEAR_PROBLEM(MODE, CONFIG)
%   [X0, G] = MALINEAR_PROBLEM('reset', MODEL, RESETCONFIG);
%   [GSTAR, RHOSTAR, LAMBDA2STAR] = MALINEAR_PROBLEM('minrobgraph', MODEL, X);


switch mode,

    % Offer basic info about the model (without creating the actual model)
    case 'info',    
        info.id = 'malinear';               % an identifier
        info.problem = @malinear_problem;   % handle to model function
        info.det = 1;                       % whether problem is deterministic or stochastic
        varargout = {info};

    case 'shortid',    
        varargout = {'malin'};               % an identifier
    
    % The 'model' mode is required
    case 'model',
        % these should stay fixed
        m.id = 'malinear';                  % an identifier
        m.problem = @malinear_problem;      % handle to m function
        m.det = 1;                          % whether problem is deterministic or stochastic        
        m.fun = 'malinear_mg';              % dynamics and rewards function
        m.plotfun = 'malinear_plot';       
        m.visualizefun = 'malinear_visualize';
        m.simagent = 'malinear_simagent';
        
        % configurable
        m.nagents = [];                     % number of agents
        m.matype = 'varconn';               % 'varconn' or 'fixedconn' problem
        m.G = [];                           % connectivity graph, for fixed connectivity
        m.range = [];                       % communication range, for variable connectivity
        m.Ts = 1;                           % sample time
        m.type = 'doubleint';               % agent type: only 'doubleint' supported so far
        m.A = [];                           % transfer matrix
        m.B = [];                           % input matrix
        m.dim = [];                         % for doubleint: dimensionality of the coordinate space
        m.rewtype = 'flock';                % flocking reward: explicit terms for all agents
        % flocking reward parameters:
        m.flock_dist = [];                  % distance cutoff (by default, the range)
        m.flock_alpha = 0.5;                % weighting of position cost term
        m.flock_beta = 0.5;                 % weighting of velocity cost term
        m.flock_maxdvel = [];               % max velocity error: beyond, rewards are insensitive
        
        % process any custom config
        if nargin >= 2, m = parseconfig(varargin{1}, m); end;
        
        switch m.matype,
            case 'varconn',
                % nothing to do
            case 'fixedconn',
                if ~strcmp(m.rewtype, 'none'), error('Type [fixedconn] should be matched with reward [none]'); end;
            otherwise,
                error('Unknown multiagent type [%s]', m.matype);
        end;
        
        switch m.type,
            case 'doubleint',
                m.A = [eye(m.dim) eye(m.dim).*m.Ts; zeros(m.dim) eye(m.dim)];
                m.B = [zeros(m.dim); eye(m.dim).*m.Ts];
            otherwise,
                error('Unknown model type [%s]', m.type);
        end;
        m.p = size(m.A, 1) + zeros(1, m.nagents);
        m.q = size(m.B, 2) + zeros(1, m.nagents);
        m.pranges = [0 cumsum(m.p)]; m.qranges = [0 cumsum(m.q)];   % variable ranges

        switch m.rewtype,
            case 'none',
                if ~strcmp(m.matype, 'fixedconn'), error('Reward [none] should be matched with type [fixedconn]'); end;
            case 'flock',
                % error checks and defaults
                if ~strcmp(m.type, 'doubleint'), 
                    error('[flock] rewards only supported for [doubleint]');
                elseif isempty(m.range) || isempty(m.flock_maxdvel),
                    error('Missing parameter for [flock] rewards'); 
                elseif m.flock_alpha + m.flock_beta ~= 1, 
                    error('flock_alpha+flock_beta should be 1'); 
                end;
                if isempty(m.flock_dist), m.flock_dist = m.range; end;
                % normalization constants
                m.flock_vnorm = m.flock_maxdvel ^ 2;
                m.flock_pnorm = (m.range - m.flock_dist) ^ 2;
            otherwise,
                error('Unknown reward type [%s]', m.rewtype);
        end;
        varargout = {m};

    case 'reset',   % handle state reset
        % WE ASSUME DOUBLE INTEGRATOR HERE
        RES.type = [];
        RES.x0 = [];
        m = varargin{1};
        res = varargin{2};
        if ischar(res) && any(strcmp(res, {'rand', 'values'})),   % handle simple method identifier
            res = struct('type', res);
        elseif isnumeric(res),
            res = struct('type', 'values', 'x0', res);
        end;
        
        res = parseconfig(res, RES);
        switch res.type,
            case 'rand',
                x0 = [rand(m.dim, m.nagents) * m.nagents * m.range/3; 
                      (2*rand(m.dim, m.nagents)-1) * m.range/20/m.Ts];
            case 'values',
                if size(res.x0, 2) > 1,    x0 = res.x0(:, unidrnd(size(res.x0, 2)));
                else                       x0 = res.x0;
                end;
        end;          
        % compute connections graph
        if m.matype(1) == 'v',  G = buildgraph(m, x0);
        else                    G = m.G;
        end;
    
        % flatten and output x, output G
        varargout = {x0(:), G};
        
    case 'minrobgraph',
        m = varargin{1}; x = varargin{2};
        % compute minimal robust graph corresponding to some given state/position config
        if (m.matype(1) ~= 'v') || (m.rewtype(1) ~= 'f'), 
            error('Minimal robust graph only makes sense in variable-connectivity flocking problems');
        end;
        [G, dp] = buildgraph(m, x);
        % DP algo as in Martin & Girard
        rho = m.range - dp;
        for k = 1:m.nagents,
            rhoold = rho;
            for i = 1:m.nagents,
                for j = 1:m.nagents,
                    rho(i,j) = max( rhoold(i, j), min(rhoold(i, k), rhoold(k, j)) );
                end;
            end;
        end;
        rhorob = min(min(rho));
        Grob = dp <= (m.range - rhorob);
        Grob(1:m.nagents+1:m.nagents*m.nagents) = 0;
        Lrob = diag(sum(Grob, 2)) - Grob; lambda = sort(eig(Lrob)); lambda2rob = lambda(2);
        varargout = {Grob, rhorob, lambda2rob};
        
end;        % mode SWITCH

end % =================== MAIN FUNCTION


% local funs
function [G, dp] = buildgraph(m, x)
if size(x, 2) == 1, x = reshape(x, m.p(1), m.nagents); end;
pos = x(1:m.dim, :);
dp = repmat(pos(:), 1, m.nagents) - repmat(pos, m.nagents, 1);
dp = squeeze(sqrt(sum(reshape(dp, m.dim, m.nagents, m.nagents).^2, 1)));
G = dp <= m.range;
G(1:m.nagents+1:m.nagents*m.nagents) = 0;     % set diagonal to 0 (no loops)
end

% END sample_problem() RETURNING out
% ===================================================