function varargout = lspiexp(cfg)
% Wrapper for batch experiments with LSPI algorithms
% Supports the following features:
%   - Running batches of experiments with various parameter values. Basic parameter values
%   that can vary are N, hN (number of state basis functions), M (number of discrete actions or
%   degree of the polynomial action approximator), Ns, hNs (number of samples). A flexible
%   mechanism is provided with which new varying parameters can be added.
%   - Evaluating batches of experiments (computing returns from sets of representative initial
%   states). A cache is used to avoid re-evaluating the same experiment twice.
%   - Displaying (at the console) or plotting statistics of the experiments: performance,
%   number of iterations to convergence, performance evolution for online experiments.
%   Statistics include number of non-convergent and oscillating experiments.
%   - Cleaning up data files (optimizing space by removing variables that are no longer
%   necessary after the experiment has completed)
%   - Plotting solutions and replaying for individual experiments. This is done by forwarding
%   the call to the LSPI function for each experiment in the batch
%
% The Q-function, and if applicable the policy, approximators can be fully configured. Data
% file names are generated automatically.
% Please see the commented config defaults for details.
% 
% Limitations:
% - Assumes hyperbox Euclidean state space with bounds in model.maxx
% - With continuous actions, only supports scalar actions

% Version: continuously changed according to need...

% TODO flexible parameter handling requires testing especially in the sample reuse code

if nargin < 1, cfg = struct; end;
% main config
CFG.run = 0;            % run experiment unless data file already exists
CFG.overwrite = 0;      % re-compute experiments even if data files already existing
CFG.stats = 0;          % statistics 
CFG.perf = 1;           % including performance by default
CFG.onlineperf = 0;     % evolution of policy performance for online LSPI
CFG.onlineconv = 0;     % online convergence
CFG.onlinereturn  = 0;  % online return
CFG.sol = 0;            % solution plots
CFG.evol = 0;           % evolution plots (with pauses, may be time consuming if many experiments)
CFG.traj = 0;
CFG.exth = [];
CFG.replay = 0;         % do a replay for each experiment
CFG.cleanup = 0;        % clean up data file (A, b matrices history)
CFG.verb = 3;
CFG.repeat = 1;         % number of experiment repetitions
CFG.startat = 1;        % start at this experiment index
CFG.miscinfo = '';
CFG.donotexpand = {};
    % do not expand these parameters; instead, they should be a vector with one value 
    % for every experiment (where the experiments are given by the expanded parameters)
% problem config
CFG.problem = 'dc_problem';
CFG.model_params = {};
% LSPI config
CFG.alg = 'lspi';       % which algorithm to use: LSPI, LSPIH, LSPIONLINE, LSPIHONLINE, LSPEONLINE
CFG.gamma = 0.95;       % discount factor(s)
CFG.eps = .01;          % threshold
CFG.maxiter = 100;      % max # iter
CFG.silent = 0;
CFG.lspi_params = struct;   % any other lspi cfg fields; can be single structure or structure array 
CFG.alpha  = 1;         % learning rate (applies to LSPEONLINE only)
% Q-function approx config
CFG.type = 'rbfdisc';   % approx type
CFG.N = -1;             % number(s) of fuzzy cores/RBF centers on each axis to try out
CFG.M = -1;             % number(s) of discrete actions; or order(s) of polynomial approximators
CFG.xspace = 'eq';      % 'eq'uidistant, 'log'arithmic
CFG.uspace = 'log';     % 'eq'uidistant, 'log'arithmic, 'pre'defined
CFG.U = [];             % action grids, if uspace='pre'
CFG.rbfrad = 'avg';     % 'avg' average distance to neigbors; 'max' max dist to neigh; vector = identical radii for all RBFs
CFG.approx_params = struct; % any other approximator cfg fields
% policy evaluation samples config
CFG.Ns = -1;            % number(s) of samples
CFG.method = 'randdisc';% samples generation method
CFG.samples_params = struct; % any other cfg fields
CFG.explor = 0;         % exploration probability in sample generation
% h approx config -- used only when alg=lspih
CFG.htype = 'rbfsing';  
CFG.happrox_params = struct;
% if left empty will assume values from the Q-fun approximator
CFG.hN =[];
CFG.hxspace = [];
CFG.hrbfrad = [];
% policy improvement samples config  -- used only when alg=lspih
CFG.hmethod = 'rand';
CFG.hNs = [];
CFG.hsamples_params = struct;
% any constraints -- used only when alg=lspih
CFG.con = [];
% online config
CFG.dupd = [];          % theta update interval
CFG.trialtime = [];     % trial time
CFG.explordecay = [];   % exploration decay rate (per second)
% other sample config (reuse, adaptive)
CFG.adaptiveNs = 0;     % adapt # of samples to # of parameters
CFG.samesamples = 0;    % use the same samples for running the algorithm w/ different settings
CFG.usesamples = [];    % configuration of experiment set to use samples from
CFG.sstest = {'N', 'M', 'Ns'};      % test these fields for loading LSTD-Q samples
CFG.hsstest = {'N', 'M', 'hNs'};    % test these fields for loading h improve samples
    % TODO this should probably be changed -- 2008-12 I don't remember what I meant by this...
% performance eval config
CFG.X0 = [];            % set of states where to evaluate return
CFG.quantize = 0;       % quantize control actions
CFG.mc_eps = 0.01;
CFG.pcache = 'd:\work\m\approxrl\data\lspipcache.mat';
CFG.forceeval = 0;      % force perf eval even for non-convergent / non-invertible experiments
CFG.forcerecompute = 0;
CFG.conftype = 'gauss'; % 'gauss', 'ttest' (note ttest is more correct, gauss kept for compatibility reasons)
CFG.verb = 5;
CFG.plot = {};
% figure config
CFG = setfigprop(CFG, 'addfields'); CFG.ylim_matrix = []; CFG.markerspace = 3;
CFG.fixedfigname = 0;   % use a fixed figure filename (no infixes, no postfixes)
CFG.Jstar = [];         % optimal score to add to score/perf figure (if any)
CFG.expprefix = '';     % prefix to add to legend entries -- describes experiment for multi-exp plots
CFG.plotminmax = 0;     % add min and max
CFG.plotconf = 0;       % add 95% confidence levels (cannot be used in conjunction with plotminmax)
CFG.minmaxtype = 'line';% 'line', 'area' -- also applies to STD plots
CFG.transparency = 1;   % whether to use transparency in shaded min-max or conf interval
CFG.showkupd = 1;      % 1 = show kupd in addition to dupd, 2 or 'only' = show ONLY kupd
% for online
CFG.opcache = 'd:\work\m\approxrl\data\lspionlpcache.mat';
CFG.donlineperf = 60;   % time interval for evaluating performance
% data files config
CFG.datadir = [];
CFG.fixeddf = '';       % if a fixed file name should be used
CFG.infix = '';
CFG.postfix = '';
CFG.hide = {};          % hide these fields from the file name
CFG.show = {'rep'};     % force showing these fields in the file name
    % Fields can include 'approx', 'N', 'M', 'samples', 'Ns', 'gamma', 'rep' and others... 
    % check the makefilename subfunction for details
% replay
CFG.x0 = [];
CFG.tend = 5;
% these fields get copied to the Qiteration config
LSPIFIELDS = {'problem', 'eps', 'maxiter', 'silent', 'model_params', 'verb'};
MCFIELDS = {'mc_eps', 'gamma', 'quantize'};

% what other fields can take multiple values, besides N, M, Ns
EXPANDABLE = {'gamma', 'explor', 'hN', 'hNs', 'dupd', 'trialtime', 'explordecay', 'alpha'};
% custom parameter names and display styles
PARNAMES.gamma = 'g'; PARSTYLES.gamma = '%.3f'; 
PARNAMES.explor = 'e'; PARSTYLES.explor = '%.1f'; 
PARNAMES.hN = 'N'; PARSTYLES.hN = '%d'; 
PARNAMES.hNs = 'Ns'; PARSTYLES.hNs = '%d'; 
PARNAMES.dupd = 'dupd'; PARSTYLES.dupd = '%.3f';
PARNAMES.trialtime = 'tt';PARSTYLES.trialtime = '%.2f';
PARNAMES.explordecay = 'ed'; PARSTYLES.explordecay = '%.4f'; 
PARNAMES.alpha = 'a'; PARSTYLES.alpha = '%.3f';

% parse and echo config
cfg = parseconfig(cfg, CFG);
dispx('LSPIEXP will run with the following config:', cfg.verb, 1);
dispx(cfg, cfg.verb, 1, 'cfg', 'config');

global LSPIEXP_STATS

% config deps, postproc, easy-to-use flags etc
if ~iscell(cfg.model_params), cfg.model_params = {cfg.model_params}; end;
% flag whether using explicit policy parameterization
cfg.uselspih = any(strcmp(cfg.alg, {'lspih', 'lspihonline'}));
% flag whether online learning
cfg.useonline = any(strcmp(cfg.alg, {'lspionline', 'lspihonline', 'lspeonline'}));
% provide hooks for possibly different parameters of the h approx (currently unused)
if cfg.uselspih,
    if isempty(cfg.hN) && cfg.uselspih, cfg.hN = cfg.N; end;
    if isempty(cfg.hxspace), cfg.hxspace = cfg.xspace; end;
    if isempty(cfg.hrbfrad), cfg.hrbfrad = cfg.rbfrad; end;
    cfg.hM = cfg.M;
end;

[cfg.expanded, cfg.flatparams, cfg.varying] = expand_params(cfg, EXPANDABLE);
cfg.nexp = size(cfg.flatparams, 2); % number of experiments

% identify some approx types
switch cfg.type,
    case {'rbfpoly','rbfortho','triangpoly','triangortho'},
        tpoly = 1;      % flag using polynomial approximator
    otherwise,
        tpoly = 0;
end;


% run experiments
if cfg.run,
    for rep = cfg.startat:(cfg.startat + cfg.repeat - 1),
        if cfg.repeat > 1, dispx(['Running set of experiments #' num2str(rep)], cfg.verb, 1); end;
        for iexp = 1:cfg.nexp,
            % retrieve parameters, form filename
            par = get_params(iexp, cfg.flatparams(:, iexp), cfg.expanded, EXPANDABLE, cfg);
            df = gen_dfname(cfg, par, cfg.expanded, EXPANDABLE, rep);
            
            % skip experiment if already ran
            if exist([df '.mat'], 'file') && ~cfg.overwrite, 
                dispx(['Skipping exp (found data file), df=' df], cfg.verb, 5);
                continue; 
            end;
            % log entry
            if cfg.repeat > 1, 
                dispx(['Running LS exp in set #' num2str(rep) ', df=' df], cfg.verb, 2);
            else
                dispx(['Running LS exp, df=' df], cfg.verb, 2);
            end;

            % Arrange the configs
            % Step 1: approx config
            acfg = struct;
            acfg.enhanced = 1;
            acfg = copyfields(cfg, acfg, {'problem', 'type', 'xspace', 'uspace', 'U', 'rbfrad'});
            acfg = copyfields(cfg.approx_params, acfg);
            acfg.N = par.N; acfg.M = par.M;
            if ~cfg.useonline,
                % Step 2: samples generation config
                scfg = struct;
                scfg = copyfields(cfg, scfg, {'method'});
                scfg.explor = par.explor;   % expandable parameter
                scfg = copyfields(cfg.samples_params, scfg);
                scfg.N = par.Ns;
            end;
            % explicit policy parameterization steps
            if cfg.uselspih,
                % Step H1: approx config -- this should be changed when policy approx
                % can be configured independently from Q-approx
                hacfg.enhanced = 1;
                hacfg.type = cfg.htype;
                hacfg.N = par.hN;
                hacfg.xspace = cfg.hxspace;
                hacfg.rbfrad = cfg.hrbfrad;
                hacfg = copyfields(cfg, hacfg, {'problem'});
                hacfg = copyfields(cfg.happrox_params, hacfg);
                % Step H2: h samples config
                hscfg = struct;
                hscfg.method = cfg.hmethod;
                hscfg.N = cfg.hNs;
                hscfg = copyfields(cfg.hsamples_params, hscfg);
            end;

            % Make experiment config
            ecfg = struct;
            % any extra parameters
            if length(cfg.lspi_params) == 1,    % same for all experiments
                ecfg = copyfields(cfg.lspi_params, ecfg);
            else                                % varying by params for current experiment
                ecfg = copyfields(cfg.lspi_params(iexp), ecfg);
            end;
            ecfg.run = 1;
            ecfg.verb = cfg.verb;
            ecfg.gamma = par.gamma;
            ecfg.datafile = df;
            ecfg.datadir = cfg.datadir;
            % set approximator configs
            ecfg.approx = acfg;
            if cfg.uselspih, ecfg.happrox = hacfg; ecfg.con = cfg.con; end;
            if cfg.useonline, 
                ecfg.dupd = par.dupd; ecfg.trialtime = par.trialtime;
                ecfg.explordecay = par.explordecay;
                if strcmp(cfg.alg, 'lspeonline'), % learning rate only applies to this alg
                    ecfg.alpha = par.alpha;
                end;
            end;
            % samples config
            if isempty(cfg.usesamples) && ~cfg.samesamples,
                % generate samples now
                if ~cfg.useonline, ecfg.samples = scfg; end;
                if cfg.uselspih, ecfg.hsamples = hscfg; end;
            elseif ~isempty(cfg.usesamples),
                % use samples from a different configuration
                scfg = parseconfig(cfg.usesamples, CFG); 
                % below some copy paste to make scfg similar to cfg
                % note sample reuse DOES NOT apply to LSPIONLINE, LSPEONLINE
                % so no need to check for that here
                scfg.uselspih = strcmp(scfg.alg, 'lspih');
                scfg.useonline = strcmp(scfg.alg, 'lspihonline');
                if isempty(scfg.hN), scfg.hN = cfg.N; end;
                if isempty(scfg.hxspace), scfg.hxspace = cfg.xspace; end;
                if isempty(scfg.hrbfrad), scfg.hrbfrad = cfg.rbfrad; end;
                [sexpanded, sfp] = expand_params(scfg, EXPANDABLE);
                if ~cfg.useonline,
                    % p.eval samples; check the sample generation method, throw a warning if not the same
                    if ~strcmp(cfg.method, scfg.method),
                        dispx(['WARNING! Current h-eval samples method [' cfg.method ...
                            '] does not match the UseSamples method [' scfg.method ']'], cfg.verb, 0);
                    end;
                    % find a matching parameters configuration on the usesamples config
                    spar = find_params(par, cfg.sstest, sfp, sexpanded, EXPANDABLE, scfg);
                    if ~isempty(spar),
                        ecfg.loadsamples = gen_dfname(scfg, spar, sexpanded, EXPANDABLE, rep);
                        dispx(['Using h-eval samples from [' ecfg.loadsamples ']'], cfg.verb, 4);
                    else
                        dispx('WARNING! Could not find h-eval sample match. Will generate new samples.', cfg.verb, 0);
                        ecfg.samples = scfg;
                    end;
                end;
                if cfg.uselspih,
                    if ~scfg.uselspih,
                        dispx('WARNING! UseSamples did not use explicit policies. Will generate new h-improve samples.', cfg.verb, 0);
                        ecfg.hsamples = hscfg;
                    else
                        % same story as above, but for the policy samples now
                        if ~strcmp(cfg.hmethod, scfg.hmethod),
                            dispx(['WARNING! Current h-improve samples method [' cfg.hmethod ...
                                '] does not match the UseSamples method [' scfg.hmethod ']'], cfg.verb, 0);
                        end;
                        spar = find_params(par, cfg.hsstest, sfp, sexpanded, EXPANDABLE, scfg);
                        if ~isempty(spar),
                            ecfg.loadhsamples = gen_dfname(scfg, spar, sexpanded, EXPANDABLE, rep);
                            dispx(['Using h-improve samples from [' ecfg.loadsamples ']'], cfg.verb, 4);
                        else
                            dispx('WARNING! Could not find h-improve sample match. Will generate new samples.', cfg.verb, 0);
                            ecfg.hsamples = hscfg;
                        end;
                    end;
                end;
            elseif cfg.samesamples,
                % find matching parameters within the same experiment; 
                % as opposed to usesamples, this time check fields for both types of samples
                tf = cfg.sstest; if cfg.uselspih, tf = {tf{:}, cfg.hsstest{:}}; end;
                [spar, sind] = find_params(par, tf, flatparams, expanded, EXPANDABLE, cfg);
                if ~isempty(spar) && sind < iexp,
                    spar = get_params(iexp, cfg.flatparams(:, fe), cfg.expanded, EXPANDABLE, cfg);
                    if ~cfg.useonline,
                        ecfg.loadsamples = gen_dfname(cfg, spar, cfg.expanded, EXPANDABLE, rep);
                    end;
                    if cfg.uselspih, 
                        ecfg.loadhsamples = gen_dfname(cfg, spar, cfg.expanded, EXPANDABLE, rep); 
                    end;
                    dispx(['Using h-eval and h-improve samples generated for [' ecfg.loadsamples ']'], cfg.verb, 4);
                else
                    % this is the 1st (or only? shouldn't happen)
                    % experiment with Ns samples, need to generate the samples
                    if ~cfg.useonline, ecfg.samples = scfg; end;
                    if cfg.uselspih, ecfg.hsamples = hscfg; end;
                end;
            else
            end;
            ecfg = copyfields(cfg, ecfg, LSPIFIELDS);

            % run experiment
            feval(cfg.alg, ecfg);

        end;    % experiments loop
    end;        % experiment repetitions loop
end;        % RUN mode


% plot various quantities or replay
% this branch is only for options that can be fully delegated to the algorithm functions
if cfg.sol || cfg.replay || cfg.evol ...
        || (cfg.useonline && cfg.traj) ...
        || (cfg.uselspih && ~isempty(cfg.exth)),
    FIGH = [];
    for iexp = 1:cfg.nexp,
        % retrieve parameters
        par = get_params(iexp, cfg.flatparams(:, iexp), cfg.expanded, EXPANDABLE, cfg);
        % make config
        ecfg = struct;
        % what to plot
        ecfg = copyfields(cfg, ecfg, {'sol', 'replay', 'evol', 'traj', 'exth'});
        if cfg.replay, ecfg = copyfields(cfg, ecfg, {'x0', 'tend'}); end;
        ecfg.verb = cfg.verb;
        % apply options for all experiments
        for rep = cfg.startat:(cfg.startat + cfg.repeat - 1),
            % form filename
            df = gen_dfname(cfg, par, cfg.expanded, EXPANDABLE, rep);
            ecfg.datafile = df;
            % call LSPI to make plots
            FIGH = [FIGH feval(cfg.alg, ecfg)];
        end;
    end;    % experiments loop
    varargout = {FIGH};
end;        % RUN mode

% compute performance of found policies & retrieve experiment stats 
if cfg.stats,
    % don't bother me with LTI update warnings 
    % when loading files made on other computers/Matlab versions...
    warning('off', 'control:LTIUpdate');
    if cfg.perf,
        X0 = feval(cfg.problem, 'X0', cfg.X0);
        ADDTOCACHE = {};
        % silence var not found warning since we rely on that to determine whether certain
        % data is cached or or not
        warning('off', 'MATLAB:load:variableNotFound');
        cmodified = 0;      % flag for modified cache
    end;

    % TODO for LSPIH this should also handle poorly conditioned HPHI matrix! & maybe other
    % things...
    % init stats
    STATS.lab = cell(1, cfg.nexp);
    STATS.J = nan(cfg.repeat, cfg.nexp);
    STATS.Jtab = cell(cfg.repeat, cfg.nexp);
    STATS.iter = nan(cfg.repeat, cfg.nexp);
    STATS.trun = nan(cfg.repeat, cfg.nexp);
    STATS.conv = nan(cfg.repeat, cfg.nexp);
    STATS.inv = nan(cfg.repeat, cfg.nexp);
    STATS.osc = nan(cfg.repeat, cfg.nexp);
    % these are the processed statistics
    STATS.sJ = nan(4, cfg.nexp); STATS.siter = nan(4, cfg.nexp); STATS.strun = nan(4, cfg.nexp);
    STATS.snoninv = nan(1, cfg.nexp); STATS.snonconv = nan(1, cfg.nexp); STATS.sosc = nan(1, cfg.nexp);

    for iexp = 1:cfg.nexp,
        % retrieve parameters
        par = get_params(iexp, cfg.flatparams(:, iexp), cfg.expanded, EXPANDABLE, cfg);
        % save label
        STATS.lab{iexp} = sprintf('N=%d,M=%d,Ns=%d', par.N, par.M, par.Ns); 
        for i = 4:length(cfg.expanded),
            STATS.lab{iexp} = [STATS.lab{iexp} ...
                sprintf([PARNAMES.(cfg.expanded{i}) '=' PARSTYLES.(cfg.expanded{i})], par.(cfg.expanded{i}))]; 
        end;
        
        for rep = cfg.startat:(cfg.startat + cfg.repeat - 1),
            % form filename, check it exists, and load data
            df = gen_dfname(cfg, par, cfg.expanded, EXPANDABLE, rep);
            if ~exist([df '.mat'], 'file'),
                error(['Data file [' df '.mat] not found.']);
            end;
            loadwithprefix(df, 'd_', {'cfg', 'model', 'approx', 'trun', 'deltah', 'conv'});
            if cfg.uselspih, loadwithprefix(df, 'd_', {'happrox', 'hdeltah'}); end;
            % load invertibility flag, if not present assumed ok
            VARINFO = whos('-file', df);
            if any(strcmp('Anoninv', {VARINFO.name})),  loadwithprefix(df, 'd_', {'Anoninv'});
            else                                        d_Anoninv = 0; 
            end;

            % save runtime and niter, if the experiment went OK and there were no
            % invertibility issues
            nrep = rep - cfg.startat + 1;   % normalized index of repetition
            STATS.inv(nrep, iexp) = ~d_Anoninv;
            if d_Anoninv && ~cfg.forceeval,   
                STATS.iter(nrep, iexp) = NaN;
                STATS.trun(nrep, iexp) = NaN;
            else
                STATS.iter(nrep, iexp) = find(~isnan(d_deltah), 1, 'last') - 1;
                STATS.trun(nrep, iexp) = d_trun;
            end;
%             if d_conv == -1, d_conv = 0; end;       % uncomment this if oscillating should be ignored
            STATS.conv(nrep, iexp) = d_conv;        % note oscillating experiments are considered convergent
            STATS.osc(nrep, iexp) = (d_conv == -1); % oscillating experiment if convergence flag is -1
            
            if cfg.perf && ~cfg.forceeval && (d_Anoninv || ~d_conv),    % perf eval does not make sense
                STATS.J(nrep, iexp) = NaN;
                STATS.Jtab{nrep, iexp} = [];
            end;
            % evaluate score if config requires it & a meaningful solution is available
            if cfg.perf && ((~d_Anoninv && d_conv) || cfg.forceeval),
                % check cache validity (if cache not exists then it's invalid by default)
                cvalid = 0;
                % form cache field postfix
                cfpostfix = ''; 
                if cfg.quantize && cfg.uselspih, cfpostfix  = [cfpostfix 'quant']; end;
                % (possibly also implement flat search later on)
                cf = [makealphanum(df) cfpostfix];
                if exist(cfg.pcache, 'file'),
                    lastwarn(''); load(cfg.pcache, cf);
                    if isempty(lastwarn),   % found cache field
                        dispx(['Found cached data for [' df ']. Checking validity.'], cfg.verb, 4);
                        c = eval(cf);
                        % firstly, the same algorithm has to be used
                        % also account for older configs without the alg field
                        cvalid = (strcmp(cfg.alg, 'lspi') && ~isfield(c.lspiexpcfg, 'alg')) ...
                            || strcmp(c.lspiexpcfg.alg, cfg.alg);
%                         disp(['After algorithm check: cvalid=' num2str(cvalid)]);
                        % LSPI & LSPIH common tests
                        % same model, same discounting -- otherwise different problems
                        % same X0 and at least as good MC error -- otherwise scores will be different
                        cvalid = cvalid && isequal(c.lspicfg.gamma, d_cfg.gamma) && isequal(c.model, d_model) ...
                                && isequal(c.lspiexpcfg.X0, cfg.X0) && (c.lspiexpcfg.mc_eps <= cfg.mc_eps);
%                         disp(['After model, gamma, X0, MC error check: cvalid=' num2str(cvalid)]);
                        % the approximator has to be the same 
                        cvalid = cvalid && ...
                            strcmp(c.lspiexpcfg.type, cfg.type) && (c.N == par.N) && (c.M == par.M);
%                         disp(['After approximator check: cvalid=' num2str(cvalid)]);
                        if ~cfg.useonline,
                            % the same sample generation method & #samples have to be the same
                            cvalid = cvalid && strcmp(c.lspiexpcfg.method, cfg.method) && (c.Ns == par.Ns);
%                             disp(['After sample check: cvalid=' num2str(cvalid)]);
                        end;
                        % LSPIH-only tests
                        if cfg.uselspih,
                            cvalid = cvalid && strcmp(c.lspiexpcfg.htype, cfg.htype) ...
                                && strcmp(c.lspiexpcfg.hmethod, cfg.hmethod) ...
                                && (c.hNs == par.hNs) && (c.hN == par.hN) ...
                                && (c.lspiexpcfg.quantize == cfg.quantize);
%                             disp(['After lspih check: cvalid=' num2str(cvalid)]);
                        end;
                        % LSPIONLINE-only tests
                        if cfg.useonline,
                            % NOTE should also check for trial time, explordecay but they
                            % should be in the filename; old cache fields inconsistent (they
                            % don't have those fields)
                            cvalid = cvalid && (c.dupd == par.dupd);
%                             disp(['After lspionline check: cvalid=' num2str(cvalid)]);
                        end;
                        if cvalid,  dispx(['Cache valid for [' df '].'], cfg.verb, 4);
                        else        dispx(['Cache invalid for [' df '].'], cfg.verb, 4);
                        end;
                    else        % variable not found in saved file
                        dispx(['Cache data not found for [' df '].'], cfg.verb, 4);
                    end;    
                else            % did not find datafile containing cache
                   dispx('Cache file not found.', cfg.verb, 4);
                end;
                if cvalid && cfg.forcerecompute,
                   dispx(['Force-recomputing cache for [' df '].'], cfg.verb, 4);
                   cvalid = 0;
                end;
                % when the cache is not valid, need to compute performance, extract data
                if ~cvalid,   
                    % clear data
                    c = struct;
                    mccfg = struct; mccfg = copyfields(cfg, mccfg, MCFIELDS);
                    % eval performance
                    dispx(['Evaluating performance for [' df ']'], cfg.verb, 2);
                    if cfg.uselspih,
                        if cfg.quantize,
                            mccfg.quantize = 1;
                            mccfg.U = d_approx.U;
                        end;
                        loadwithprefix(df, 'd_', {'htheta'});
                        [c.J, c.Jtab] = mc_approx(d_model, d_happrox, d_htheta, X0, mccfg);
                    else
                        loadwithprefix(df, 'd_', {'theta'});
                        % revise approx to make sure it conforms to latest requirements
                        [c.J, c.Jtab] = mc_approx(d_model, revise_approx(d_approx), d_theta, X0, mccfg);
                    end;
                    % save stats
                    c.trun = d_trun;
                    c.iter = find(~isnan(d_deltah), 1, 'last') - 1;

                    % remove samples & approx config fields, they may take a lot of space
                    % save various data in the cache field
                    d_cfg = rmfield(d_cfg, {'approx'});
                    if ~cfg.useonline, d_cfg = rmfield(d_cfg, {'samples'}); end;
                    if cfg.uselspih, d_cfg = rmfield(d_cfg, {'happrox', 'hsamples'}); end;
                    c.lspicfg = d_cfg; c.lspiexpcfg = cfg;
                    c.datafile = df;
                    c.model = d_model;
                    c.approx = d_approx; c.N = par.N; c.M = par.M; c.Ns = par.Ns;
                    c.X0 = X0;
                    if cfg.uselspih,
                        c.happrox = d_happrox; c.hNs = par.hNs; 
                        % future-proof: provide for separate M for policy approx
                        c.hN = par.hN; c.hM = par.M;
                    end;
                    if cfg.useonline, 
                        c.dupd = par.dupd; c.trialtime = par.trialtime; 
                        c.explordecay = par.explordecay;
                    end;
                    % create variable
                    eval([cf '=c;']);
                    % remember we need to save this variable on the cache
                    ADDTOCACHE{end+1} = cf;
                    cmodified = 1;
                end;        % cache not valid IF
                
                % save computed score
                STATS.J(nrep, iexp) = c.J;
                STATS.Jtab{nrep, iexp} = c.Jtab;
                
            end;        % performance evaluation required IF
        end;            % exp repetitions FOR
        
        % process stats preparing for plot/display etc
        conv = STATS.conv(:, iexp); inv = STATS.inv(:, iexp); osc = STATS.osc(:, iexp);
        if cfg.forceeval, forced = true(cfg.repeat, 1); else forced = false(cfg.repeat, 1); end;
        STATS.snoninv(iexp) = sum(~inv); 
        STATS.snonconv(iexp) = sum(inv & ~conv);    % don't count noninvertible here
        STATS.sosc(iexp) = sum(osc);
        % only collect meaningful data (invertible and convergent)
        J = STATS.J(inv & conv | forced, iexp); 
        iter = STATS.iter(inv & conv | forced, iexp); 
        trun = STATS.trun(inv & conv | forced, iexp);
        if ~any(inv & conv) && ~cfg.forceeval,
            STATS.sJ(:, iexp) = NaN; STATS.siter(:, iexp) = NaN; STATS.strun(:, iexp) = NaN;
        else
            nn = size(J, 1);    % number of valid experiments
            switch cfg.conftype,
                case 'gauss',
                    STATS.sJ(:, iexp) = [mean(J) min(J) max(J) 1.96*std(J)/sqrt(nn)];
                    STATS.siter(:, iexp) = [mean(iter) min(iter) max(iter) 1.96*std(iter)/sqrt(nn)]; 
                    STATS.strun(:, iexp) = [mean(trun) min(trun) max(trun) 1.96*std(trun)/sqrt(nn)]; 
                case 'ttest',
                    % Test if any values different from the mean to avoid division-by-zero in ttest
                    % J may be NaN if it was not computed
                    if any(isnan(J)) || all(mean(J) == J), confint = [mean(J) mean(J)]; 
                    else [hdummy, pdummy, confint] = ttest(J);
                    end;
                    STATS.sJ(:, iexp) = [mean(J) min(J) max(J) diff(confint)./2];
                    if all(mean(iter) == iter), confint = [mean(iter) mean(iter)]; 
                    else [hdummy, pdummy, confint] = ttest(iter);
                    end;
                    STATS.siter(:, iexp) = [mean(iter) min(iter) max(iter) diff(confint)./2]; 
                    if all(mean(trun) == trun), confint = [mean(trun) mean(trun)]; 
                    else [hdummy, pdummy, confint] = ttest(trun);
                    end;
                    STATS.strun(:, iexp) = [mean(trun) min(trun) max(trun) diff(confint)./2]; 
            end;
        end;
%         CONV = conv'
%         INV = inv'
%         OSC = osc'
%         TAKEN = inv' & conv' | forced'
        
    end;                % experiments FOR
    
    if cfg.perf,
        % restore var not found warning
        warning('on', 'MATLAB:load:variableNotFound');
        % re-save cache if evaluated performance, and cache was modified
        if cmodified, 
            % save non-compressed (v6) for fast loading
            % append only new variables if cache already exists
            if exist(cfg.pcache, 'file'),
                dispx('Saving new data to cache... ', cfg.verb, 2);
                save(cfg.pcache, ADDTOCACHE{:}, '-append', '-v6'); 
            else
                dispx('Creating cache... ', cfg.verb, 2);
                save(cfg.pcache, ADDTOCACHE{:}, '-v6'); 
            end;
            dispx([8 'done.'], cfg.verb, 2);
        end;
    end;

    S = cell(cfg.nexp+1, 4);
    % make a table header with useful approx and samples info
    S{1,1} = [cfg.type '|' cfg.xspace '|' cfg.uspace '; ' cfg.method];
    if ~isempty(cfg.infix), S{1,1} = [S{1,1} '/' cfg.infix]; end;
    S{1,2} = 'Perf'; S{1,3} = 'Iter'; S{1,4} = 'Exec.time';
    % precision of score display is given my the precision in evaluating it
    precscore = ['%.' num2str(ceil(log10(1/cfg.mc_eps))) 'f'];
    precruntime = '%.1f';
%     precscore = '%.0f'; precruntime = '%.0f';
    for iexp = 1:cfg.nexp,
        % put data into stats cell array for displaying
        S{iexp+1,1} = STATS.lab{iexp};
        if cfg.repeat > 1,
            if STATS.snoninv(iexp) > 0, S{iexp+1,1} = [S{iexp+1,1} '(' num2str(STATS.snoninv(iexp)) 'NI)']; end;
            if STATS.snonconv(iexp) > 0, S{iexp+1,1} = [S{iexp+1,1} '(' num2str(STATS.snonconv(iexp)) 'NC)']; end;
            if STATS.sosc(iexp) > 0, S{iexp+1,1} = [S{iexp+1,1} '(' num2str(STATS.sosc(iexp)) 'O)']; end;
            if (STATS.snoninv(iexp) + STATS.snonconv(iexp) == cfg.repeat) && ~cfg.forceeval,
                S{iexp+1,2} = '---'; S{iexp+1,3} = '---'; S{iexp+1,4} = '---';
            else
                % 4 and 5th elements display confidence interval
                S{iexp+1,2} = sprintf([precscore '(' precscore ',' precscore '; ' precscore ',' precscore ')'], ...
                    STATS.sJ(1, iexp), STATS.sJ(2, iexp), STATS.sJ(3, iexp), ...
                    STATS.sJ(1, iexp)-STATS.sJ(4, iexp), STATS.sJ(1, iexp)+STATS.sJ(4, iexp));
                S{iexp+1,3} = sprintf('%3.1f (%3d, %3d)', ...
                    STATS.siter(1, iexp), STATS.siter(2, iexp), STATS.siter(3, iexp));
                % display only mean for runtimes, for now
                % S{iexp+1,4} = sprintf('~%8.1f s', STATS.strun(1, iexp));
                S{iexp+1,4} = sprintf([precruntime '(' precruntime ',' precruntime '; ' precruntime ',' precruntime ')'], ...
                    STATS.strun(1, iexp), STATS.strun(2, iexp), STATS.strun(3, iexp), ...
                    STATS.strun(1, iexp)-STATS.strun(4, iexp), STATS.strun(1, iexp)+STATS.strun(4, iexp));
            end;
        else
            if STATS.snoninv(iexp), S{iexp+1,1} = [S{iexp+1,1} '(NI)']; end;
            if STATS.snonconv(iexp), S{iexp+1,1} = [S{iexp+1,1} '(NC)']; end;
            if STATS.sosc(iexp), S{iexp+1,1} = [S{iexp+1,1} '(O)']; end;
            if (STATS.snoninv(iexp) || STATS.snonconv(iexp))  && ~cfg.forceeval,
                S{iexp+1,2} = '---'; S{iexp+1,3} = '---'; S{iexp+1,4} = '---';
            else
                S{iexp+1,2} = sprintf(precscore, STATS.sJ(1, iexp));
                S{iexp+1,3} = sprintf('%3d', STATS.siter(1, iexp));
                S{iexp+1,4} = sprintf([precruntime ' s'], STATS.strun(1, iexp));
            end;
        end;
    end;
    
    % display log
    format long
    dispx(S, cfg.verb, 0);
    format short

    % ===========================
    % plot stats
    if ~isempty(cfg.plot),
        % style setup
        % primary color depends on whether standalone or adding to current plot
        if cfg.addtocrt, bcol = [0 0 0]; else bcol = [.45 .45 .45]; end;
        sty.avg = {'-', 'Color', bcol, 'Marker', 'o', 'MarkerSize', 5, 'MarkerEdgeColor', 'none', 'MarkerFaceColor', bcol, 'LineWidth', 1.5};
        sty.lower = {'--', 'Color', bcol, 'Marker', 'o', 'MarkerSize', 5, 'MarkerEdgeColor', 'none', 'MarkerFaceColor', bcol, 'LineWidth', 1.5};
        sty.upper = {'-.', 'Color', bcol, 'Marker', 'o', 'MarkerSize', 5, 'MarkerEdgeColor', 'none', 'MarkerFaceColor', bcol, 'LineWidth', 1.5};
        sty.conf = {'--', 'Color', bcol, 'LineWidth', 1};  % changed 2009-02-13 from sty.lower
        sty.ninv = {'^', 'Color', 'k'}; sty.nconv = {'v', 'Color', 'k'};
        sty.opt = {'-', 'Color', 'k', 'LineWidth', 1.5};
        if cfg.transparency,    sty.patch = {bcol, 'edgecolor' , 'none', 'facealpha', 0.3};
        else % use a patch with lighter color
            sty.patch = {max([.6 .6 .6], bcol + [.3 .3 .3]), 'edgecolor' , 'none'};
        end;
        % color map is black to dirty white
        sty.cm = gray(96); sty.cm = sty.cm(24:end);
        % other styles changes for addtocrt
        if cfg.addtocrt,    
            sty.avg = {'-', 'Color', bcol, 'Marker', 's', 'MarkerSize', 6, 'MarkerFaceColor', 'none'};
            sty.lower = {':', 'Color', bcol, 'Marker', 's', 'MarkerSize', 6, 'MarkerFaceColor', 'none'};
            sty.upper = {'-.', 'Color', bcol, 'Marker', 's', 'MarkerSize', 6, 'MarkerFaceColor', 'none'};
            sty.conf = {':', 'Color', bcol}; % changed 2009-02-13 from sty.lower
            sty.ninv = {'+', 'Color', 'k'}; sty.nconv = {'x', 'MarkerSize', 9, 'Color', 'k'};
        end;
        % labels
        % latexprop = {'Interpreter', 'LaTeX', 'FontSize',13};
        lab.Jpostfix = 'score'; lab.iterpostfix = 'number of iterations'; lab.trunpostfix = 'execution time';
        lab.J = 'Score'; lab.iter = 'Number of iterations'; lab.trun = 'Execution time';
        lab.trununits = 's';
        lab.avgpref = 'mean ';
        lab.minpref = 'min '; lab.maxpref = 'max '; lab.stdpref = 'std ';
        lab.opt = 'optimal score';
        lab.nconv = 'non-convergent'; lab.ninv = 'A non-invertible'; lab.osc = 'oscillating';
        lab.N = 'N'; lab.M = 'M'; lab.Ns = 'N_0'; lab.hNs = 'N_0^h'; lab.hN = 'N^h';
        lab.gamma = '\gamma'; lab.explor = '\epsilon'; lab.dupd = 'T_{\theta}';
        lab.trialtime = 'T_{trial}'; lab.explordecay = '\epsilon_{d}';
        lab.kupd = 'K_{\theta}';
        lab.trialtimeunits = 's'; lab.dupdunits = 's';
        
        % some things are different for polynomial approximators
        if tpoly, lab.M = 'M_p+1'; end;
        
        FIGH = [];
        % adaptive number of samples doesn't mean that Ns or hNs is a varying quantity that we
        % should plot against
        if cfg.adaptiveNs,
            adaptiveind = strcmp(cfg.varying, 'Ns');
            adaptiveind = adaptiveind | strcmp(cfg.varying, 'hNs');
            varying = cfg.varying(~adaptiveind);
        else varying = cfg.varying;
        end;
        if (length(varying) == 1) || (isempty(varying) && any(cfg.addtocrt)),        % 2-D plot
            % check if we're plotting a single value across an existing axis (with
            % addtocrt) or we're plotting multiple values
            if length(varying) == 1,
                single = 0;
                axis1label = varying{1}; 
                % horiz M axis for poly approx is shifted
                if tpoly && strcmp(axis1label, 'M'), 
                    axis1 = cfg.(axis1label) + 1;
                elseif strcmp(axis1label, 'dupd') && any(cfg.showkupd), % use K_theta axis instead of T_theta
                    axis1 = cfg.(axis1label) / d_model.Ts;
                    
                else axis1 = cfg.(axis1label);
                end;
                if cfg.forceeval, validexp = 1:cfg.nexp;    % all experiments will be plotted
                else validexp = find(STATS.snoninv + STATS.snonconv < cfg.repeat);
                end;
            else
                single = 1;
            end;
            for iplot = 1:length(cfg.plot),
                % init stuff depending on whether we're creating new plot or adding to current
                if any(cfg.addtocrt),        % retrieve current figure handle and legend
                    if ~islogical(cfg.addtocrt) && length(cfg.addtocrt) >= iplot ...
                            && any(findall(0) == cfg.addtocrt(iplot)) ...
                        % cfg.plot(iplot) is probably a handle to a figure, possibly different
                        % from the current one; plot in that figure
                        FIGH(end+1) = cfg.addtocrt(iplot); figure(FIGH(end));
                    else    % cfg.plot is just a flag
                        FIGH(end+1) = gcf;
                    end;
                    [legh, objh, lh, leg] = legend; clear legh objh;
                    avgsty = sty.avg; uppersty = sty.upper; lowersty = sty.lower; 
                    confsty = sty.conf; patchsty = sty.patch;
                else                    % create new figure
                    FIGH(end+1) = figure; lh = []; leg = {};
                    avgsty = sty.avg; uppersty = sty.upper; lowersty = sty.lower; 
                    confsty = sty.conf; patchsty = sty.patch;
                end;
                quant = cfg.plot{iplot};    % what quantity to plot
                % for easy coding below, determine plot data here
                if single,
                    xv = get(gca, 'XLim');
                    yv = [0 0] + STATS.(['s' quant])(1, 1);
                    yvmin = [0 0] + STATS.(['s' quant])(2, 1);
                    yvmax = [0 0] + STATS.(['s' quant])(3, 1);
                    yvconf = [0 0] + STATS.(['s' quant])(4, 1);
                else
                    xv = axis1(validexp);
                    yv = STATS.(['s' quant])(1, validexp);
                    yvmin = STATS.(['s' quant])(2, validexp);
                    yvmax = STATS.(['s' quant])(3, validexp);
                    yvconf = STATS.(['s' quant])(4, validexp);
                end;
                
                % set the limits if given separately for every plot
                if ~isempty(cfg.ylim_matrix), cfg.ylim = cfg.ylim_matrix(iplot, :); end;
                % mean is alwas plotted
                lh(end+1) = cfg.plotfun(xv, yv, avgsty{:});
                hold on;
                if cfg.repeat == 1,     % single-experiment legend
                    leg{end+1} = [cfg.expprefix ' ' lab.([quant 'postfix'])];
                elseif cfg.repeat > 1,  % plot bounds, if requested
                    leg{end+1} = [cfg.expprefix ' ' lab.avgpref lab.([quant 'postfix'])]; % include infix
                    if cfg.plotconf,
                        if strcmp(cfg.minmaxtype, 'area'),  % patches (with negative Z)
                            patch([xv fliplr(xv)], [yv-yvconf fliplr(yv+yvconf)], -.1 * [xv xv], patchsty{:});
                            % the stuff below doesn't work properly (turns back to linear
                            % axes)
%                             if any(strcmp(func2str(cfg.plotfun), {'semilogx', 'semilogy', 'loglog'})),
%                                 % need to explicitly set renderer to avoid bug
%                                 % where transparency is lost
%                                 set(gcf,'RendererMode','auto');
%                             end;
                        end;
                        lh(end+1) = cfg.plotfun(xv, yv-yvconf, confsty{:});
                        cfg.plotfun(xv, yv+yvconf, confsty{:});
                        leg{end+1} = '95% confidence bounds';
                    elseif cfg.plotminmax,
                        if strcmp(cfg.minmaxtype, 'area'),
                            % add patches for area style( with negative Z)
                            patch([xv fliplr(xv)], [yvmin fliplr(yvmax)], -.1 * [xv xv], patchsty{:});
                            % note unlike the std plot, we also add the min and max boundaries
                        end;
                        lh(end+1) = cfg.plotfun(xv, yvmin, lowersty{:});
                        lh(end+1) = cfg.plotfun(xv, yvmax, uppersty{:});
                        % add legend entries for min and max boundaries
                        leg{end+1} = [cfg.expprefix ' ' lab.minpref lab.([quant 'postfix'])];
                        leg{end+1} = [cfg.expprefix ' ' lab.maxpref lab.([quant 'postfix'])];
                    end;
                    % else, don't add anything (no std, no minmax)
                end;
                if strcmp(quant, 'J') && ~isempty(cfg.Jstar),	% add optimal score
                    lh(end+1) = cfg.plotfun(xv, cfg.Jstar+0*xv, sty.opt{:});
                    leg{end+1} = lab.opt;
                end;
                
                % signal non-convergence or non-invertibility for iter plots
                if strcmp(quant, 'iter') % && ~any(cfg.addtocrt),  % only reliable for first plot
                    % get the max Y coord on the graph either from the data or from the limits
                    % plot markers for noninv and nonconv experiments
                    ninvh = 0; nconvh = 0; osch = 0; 
                    maxmarkers = max(STATS.snonconv + STATS.snoninv + STATS.sosc);
                    if ~isempty(cfg.ylim),
                        maxy = max(cfg.ylim(2), max(max(STATS.(['s' quant]))) + cfg.markerspace * maxmarkers);
                        cfg.ylim(2) = maxy;
                    else
                        maxy = max(max(STATS.(['s' quant]))) + cfg.markerspace * maxmarkers;
                    end;
                    for iexp = 1:cfg.nexp,
                        ninv = STATS.snoninv(iexp); nconv = STATS.snonconv(iexp); osc = STATS.sosc(iexp);
                        if ninv,    ninvh = cfg.plotfun(repmat(axis1(iexp), ninv, 1), ...
                                        maxy-cfg.markerspace*(0:ninv-1), sty.ninv{:});
                        end;
                        if nconv,   nconvh = cfg.plotfun(repmat(axis1(iexp), nconv, 1), ...
                                        maxy-cfg.markerspace*(ninv:1:ninv+nconv-1), sty.nconv{:});
                        end;
                        if osc,     osch = cfg.plotfun(repmat(axis1(iexp), osc, 1), ...
                                        maxy-cfg.markerspace*(ninv+nconv:1:ninv+nconv+osc-1), sty.nconv{:}); 
                                    % should be instead: sty.osc{:});
                        end;
                    end;
                    % add nonconv and noninv to legend, if they were used
                    if ninvh ~= 0, lh(end+1) = ninvh; leg{end+1} = [cfg.expprefix ' ' lab.ninv]; end;
                    if nconvh ~= 0, lh(end+1) = nconvh; leg{end+1} = [cfg.expprefix ' ' lab.nconv]; end;
                    if osch ~= 0, lh(end+1) = osch; leg{end+1} = [cfg.expprefix ' ' lab.osc]; end;
                end;
                % set the legend, labels, etc.
                lh = legend(lh, leg{:}); set(lh, 'Location', 'Best', 'FontSize', 8);
                if ~any(cfg.addtocrt), 
                    % check if we should use kupd instead of dupd
                    if strcmp(axis1label, 'dupd') & cfg.showkupd, 
                        axis1label2 = 'kupd';
                    else axis1label2 = axis1label;
                    end;
                    if isfield(lab, [axis1label2 'units']),
                        xlabel([lab.(axis1label2) ' [' lab.([axis1label2 'units']) ']']); 
                    else
                        xlabel(lab.(axis1label2)); 
                    end;
                    if isfield(lab, [quant 'units']),
                        ylabel([lab.(quant) ' [' lab.([quant 'units']) ']']); 
                    else
                        ylabel(lab.(quant)); 
                    end;
                end;
                % does this break something? I moved it out of the addtocrt if
                setfigprop(cfg); 
                % make sure all the data is in the plot
                ylim0 = get(gca, 'ylim');
%                 ylim0(1) = min( ylim0(1), min(min(STATS.(['s' quant]))) ); 
%                 ylim0(2) = max( ylim0(2), max(max(STATS.(['s' quant]))) ); 
                set(gca, 'ylim', ylim0);
                % else figure config remains the same
                set(gcf, 'NumberTitle', 'off', 'Name', [get(gcf, 'Name') ' ' S{1,1}]);
                if cfg.addtocrt == 0,
                    % if addtocrt is explicitly set to 0, interpret it as a sign that things will
                    % be added to the figure; therefore, don't save yet
                else
%                     if any(cfg.addtocrt),   % don't add the miscinfo / exp type, multiple exp-s here
%                         fpath = [cfg.savedir cfg.savefig '_' quant];
%                     else
                    if cfg.fixedfigname,
                        fpath = [cfg.savedir cfg.savefig];
                    else
                        if isempty(cfg.miscinfo),
                            fpath = [cfg.savedir cfg.savefig '_' quant];
                        else
                            fpath = [cfg.savedir cfg.savefig '_' cfg.miscinfo '_' quant];
                        end;
                    end;
%                     end;
                    saveplot(FIGH(end), fpath, cfg.plottarget);
                end;
            end;    % plot quantity FOR
        elseif isempty(varying),
            dispx('No varying parameters and cannot add to current figure. Skipping plot', cfg.verb, 1);
        else
            dispx('WARNING! LSPIEXP can only plot against 1 varying parameter. Skipping plot', cfg.verb, 0);
        end;
    end;
    if ~isempty(cfg.plot),
        varargout = {FIGH, STATS, S};
    else 
        varargout = {STATS, S};
    end;
    
    LSPIEXP_STATS = STATS;
end;


% compute performance of found policies & retrieve experiment stats 
% TODO this should also implement a check for diverging experiments
if cfg.onlineperf || cfg.onlinereturn || cfg.onlineconv,
    % don't bother me with LTI update warnings for different Matlab versions
    warning('off', 'control:LTIUpdate');
    if cfg.onlineperf,  % initialize performance evaluation
        X0 = feval(cfg.problem, 'X0', cfg.X0);
        ADDTOCACHE = {};
        cmodified = 0;      % flag for modified cache
        % silence var not found warning -- we rely on that to check if data cached
        warning('off', 'MATLAB:load:variableNotFound');
    end;

    % readable labels
    lab.Jpostfix = 'score'; lab.retpostfix = 'return';
    lab.J = 'Score'; lab.ret = 'Return with exploration';
%     lab.deltah = {'$\Vert\theta_\ell - \theta_{\ell-1}\Vert$', 'Interpreter', 'latex', 'FontSize', 13};
    lab.deltah = {'Parameter difference'};
    lab.Jfile = 'onlinescore'; lab.deltahfile = 'deltatheta'; lab.retfile = 'onlinereturn';
    lab.avgpref = 'mean '; lab.minpref = 'min '; lab.maxpref = 'max '; lab.stdpref = 'std ';
    lab.time = 't [s]'; 
    lab.dupd = 'T_{\theta}'; lab.dupdunits = 's'; 
    lab.kupd = 'K_{\theta}'; lab.kupdunits  = ''; 
    lab.explordecay = '\epsilon_{d}'; lab.explordecayunits = ''; 
    lab.trialtime = 'T_{trial}'; lab.trialtimeunits  = 's'; 
    lab.alpha = '\alpha'; lab.alphaunits  = ''; 
    
    % init stats
    STATS.lab = cell(1, cfg.nexp);
    STATS.J = cell(1, cfg.nexp);        STATS.Jtime = cell(1, cfg.nexp);
    STATS.ret = cell(1, cfg.nexp);      STATS.rettime = cell(1, cfg.nexp);
    STATS.deltah = cell(1, cfg.nexp);   STATS.deltahtime = cell(1, cfg.nexp);
    
    % experiments loop
    for iexp = 1:cfg.nexp,
        % retrieve parameters
        par = get_params(iexp, cfg.flatparams(:, iexp), cfg.expanded, EXPANDABLE, cfg);

        STATS.J{iexp} = []; STATS.Jtime{iexp} = [];    % init stats for current experiment
        STATS.deltah{iexp} = []; STATS.deltahtime{iexp} = [];
        
        for rep = cfg.startat:(cfg.startat + cfg.repeat - 1),
            % form filename, check it exists, and load data from that file
            df = gen_dfname(cfg, par, cfg.expanded, EXPANDABLE, rep);
%             disp(df);
            if ~exist([df '.mat'], 'file'), error(['Data file [' df '.mat] not found.']); end;
            loadwithprefix(df, 'd_', {'cfg', 'model', 'approx', ...
                'trun', 'deltah', 'updatesk', 'nupd', ...
                'ntr', 'X', 'R', 'Ktrial'});      % trial data
            
            % add stats (they are nearly for free so we add them even if we don't use them)
            STATS.deltah{iexp}(rep-cfg.startat+1, 1:d_nupd-1) = d_deltah(2:d_nupd);
            STATS.deltahtime{iexp}(rep-cfg.startat+1, 1:d_nupd-1) = d_updatesk(2:d_nupd) * d_model.Ts;
            if size(STATS.deltah{iexp}, 1) > 1 && any(any( diff(STATS.deltahtime{iexp}, 1, 1) )),
                warning('APPROXRL:warning', 'Not all experiments updated policy theta at the same time intervals.');
            end;

            if cfg.onlineperf,      % evaluate performance (or load from cache)
                % form vector of times where theta should be evaluated
                if isscalar(cfg.donlineperf),   % equidistant eval times
                    evaltimes = cfg.donlineperf : cfg.donlineperf : d_cfg.maxtime;
                else evaltimes = cfg.donlineperf;
                end;
                % try using cache (check validity)
                cvalid = 0; 
                % form cache field postfix
                cfpostfix = ''; 
                if cfg.quantize && cfg.uselspih, cfpostfix  = [cfpostfix 'quant']; end;
                cf = [makealphanum(df) cfpostfix];
                if exist(cfg.opcache, 'file'),
                    lastwarn(''); load(cfg.opcache, cf);
                    if isempty(lastwarn),   % found cache field
                        dispx(['Found cached data for [' df ']. Checking validity.'], cfg.verb, 4);
                        c = eval(cf);
                        % the same algorithm has to be used
                        cvalid = strcmp(cfg.alg, c.lspiexpcfg.alg);
                            % disp(sprintf('After algorithm type check, cvalid=%d', cvalid));
                        % same model, same discounting, same X0 and at least as good MC error
                        % evaluate performance as often
                        % model check disabled due to incompatibilities with Linux
                        % experiments...
                        %  && isequal(c.model, d_model) ...
                        cvalid = cvalid && isequal(c.lspicfg.gamma, d_cfg.gamma) ...
                                && isequal(c.lspiexpcfg.X0, cfg.X0) && (c.lspiexpcfg.mc_eps <= cfg.mc_eps);
                                % disp(sprintf('After common parameters check, cvalid=%d', cvalid));
                        % check all eval times were evaluated in old cache, too
                        for i = 1:length(evaltimes),
                            cvalid = cvalid && any(evaltimes(i) == c.evaltimes);
                        end;
                                % disp(sprintf('After evaltimes check, cvalid=%d', cvalid));
                        % same approximator
                        cvalid = cvalid && strcmp(c.lspiexpcfg.type, cfg.type) && (c.N == par.N) && (c.M == par.M);
                                % disp(sprintf('After approximator check, cvalid=%d', cvalid));
                        % LSPIH-only tests
                        if cfg.uselspih,    
                            % there should also be more tests here (h approx, h samples)
                            % everything should be in cache field name so that it saves us, though...
                            % cvalid = (c.lspiexpcfg.quantize == cfg.quantize);
                        end;
                        % same online parameters
                        % NOTE should also check for trial time, explordecay... 
                        % (assuming they are in the file name, it should be fine)
                        cvalid = cvalid && (c.dupd == par.dupd);
                                % disp(sprintf('After online parameters check, cvalid=%d', cvalid));
                        if cvalid,  dispx(['Cache valid for [' df '].'], cfg.verb, 4);
                        else        dispx(['Cache invalid for [' df '].'], cfg.verb, 4);
                        end;
                    else    dispx(['Cache data not found for [' df '].'], cfg.verb, 4);
                    end;    
                else    dispx('Cache file not found.', cfg.verb, 4);
                end;
                if cvalid && cfg.forcerecompute, 
                   dispx(['Force-recomputing cache for [' df '].'], cfg.verb, 4);
                   cvalid = 0;
                end;
                % when the cache is not valid, need to compute performance, extract data
                if ~cvalid,
                    % clear data
                    c = struct; 
                    % eval performance
                    dispx(['Evaluating performance for [' df ']'], cfg.verb, 2);
                    if cfg.uselspih,
                        loadwithprefix(df, 'd_', {'happrox', 'hthetah', 'thetahk', 'Ts'});
                        ntheta = length(d_hthetah);
                    else
                        loadwithprefix(df, 'd_', {'thetah', 'thetahk', 'Ts'});
                        ntheta = length(d_thetah);
                    end;
                    mccfg = struct; mccfg = copyfields(cfg, mccfg, MCFIELDS);
                    c.J = []; c.Jtime = []; ieval = 1;
                    for ith = 1:ntheta,
                        if (d_thetahk(ith)*d_Ts) >= evaltimes(ieval)
                            dispx(['Evaluating theta at time ' num2str(d_thetahk(ith)*d_Ts)], cfg.verb, 5);
                            if cfg.uselspih,
                                if cfg.quantize,
                                    mccfg.quantize = 1;
                                    mccfg.U = d_approx.U;
                                end;
                                c.J(end+1) = mc_approx(d_model, d_happrox, d_hthetah{ith}, X0, mccfg);
                            else
                                c.J(end+1) = mc_approx(d_model, d_approx, d_thetah{ith}, X0, mccfg);
                            end;
                            c.Jtime(end+1) = d_thetahk(ith) * d_Ts;
                            ieval = ieval + 1;
                            if ieval > length(evaltimes), break; end;
                        end;
                    end;
                    % save various other data
                    c.evaltimes = evaltimes;
                    c.trun = d_trun; c.iter = find(~isnan(d_deltah), 1, 'last') - 1;    % crtly unused
                    c.lspicfg = d_cfg; c.lspiexpcfg = cfg;
                    c.datafile = df; c.model = d_model; c.approx = d_approx; c.N = par.N; c.M = par.M; 
                    c.X0 = X0;
                    c.dupd = par.dupd;
                    c.trialtime = par.trialtime;
                    c.explordecay = par.explordecay;
                    c.alpha = par.alpha;
                    % create cache variable
                    eval([cf '=c;']);
                    % remember we need to save this data on the cache
                    ADDTOCACHE{end+1} = cf; cmodified = 1;
                end;        % cache not valid IF

                % save computed (or retrieved from cache) score
                STATS.J{iexp}(rep-cfg.startat+1, 1:length(c.J)) = c.J;
                STATS.Jtime{iexp}(rep-cfg.startat+1, 1:length(c.Jtime)) = c.Jtime;
            end;        % evaluate performance IF
            
            if cfg.onlinereturn,    % compute return
                % Here we assume continuing tasks, fixed-length trials
                switch d_cfg.reset,
                    case 'rand',        error('reset=rand, comparing returns is meaningless');
                    case 'fixed',       trialrange = 1:d_ntr;                   % use all data
                    case 'randfixed',   trialrange = 1:d_cfg.resetper:d_ntr;    % use only these trials
                    otherwise,          error('Unsupported reset type [%s]', d_cfg.reset);
                end;
                X0 = reshape(d_X(:, 1, trialrange), [], length(trialrange));
                R = reshape(d_R(1, :, trialrange), [], length(trialrange));
                % check if all the trials really start in the same x0
                if any(diff(X0, 1, 2)), error('Computing returns: The initial states are not the same.'); end;
                % compute time indices when the picked trials started
                STATS.rettime{iexp}(rep-cfg.startat+1, 1:length(trialrange)) = (trialrange - 1) .* d_Ktrial .* d_model.Ts;
                discounting = repmat(d_cfg.gamma .^ (0:d_Ktrial-1)', 1, length(trialrange));
                STATS.ret{iexp}(rep-cfg.startat+1, 1:length(trialrange)) = sum(discounting .* R(2:end, :));
%                 rettime = STATS.rettime{iexp}
%                 ret = STATS.ret{iexp}
%                 pause;
            end;

        end;            % experiment repetitions FOR
        
        if size(STATS.J{iexp}, 1) > 1 && any(any( diff(STATS.Jtime{iexp}, 1, 1) )),
            warning('APPROXRL:warning', 'Not all experiments saved theta at the same time intervals.');
        end;

        % save experiment label
        STATS.lab{iexp} = ''; % sprintf('N=%d,M=%d,', par.N, par.M); 
        for i = 4:length(cfg.expanded),
            STATS.lab{iexp} = [STATS.lab{iexp} ...
                sprintf(['%s=' PARSTYLES.(cfg.expanded{i}) '%s'], ...
                lab.(cfg.expanded{i}), par.(cfg.expanded{i})), lab.([cfg.expanded{i} 'units'])]; 
            if strcmp(cfg.expanded{i}, 'dupd'),
                if cfg.showkupd == 1,
                    % add the meaning of the parameter in # samples
                    STATS.lab{iexp} = [STATS.lab{iexp} ....
                        sprintf(' (%s=%d)', lab.kupd, par.(cfg.expanded{i}) / d_model.Ts)]; 
                elseif strcmp(cfg.showkupd, 'only') || (isscalar(cfg.showkupd) && (cfg.showkupd == 2)),
                    % use only that meaning
                    STATS.lab{iexp} = sprintf('%s=%d', lab.kupd, par.(cfg.expanded{i}) / d_model.Ts); 
                end;
            end;
        end;
    
    end;                % experiments FOR
    
    if cfg.onlineperf,  % finalize cache
        % restore var not found warning
        warning('on', 'MATLAB:load:variableNotFound');
        % re-save cache if evaluated performance, and cache was modified
        % append only new variables; save non-compressed (v6) for fast loading
        if cmodified,         
            disp('Adding new information to (online LS) cache');
            save(cfg.opcache, ADDTOCACHE{:}, '-append', '-v6'); 
        end;
    end;
    
    % styles (generated programatically)
    % note these colors have to be given in RGB to work with the
    % no-transparency shading option
    cs.colors = {'k', 'r', 'b', 'g', 'm', 'c', [.5 .5 .5], 'r'};
    gs.colors = {[0 0 0], [.4 .4 .4], [.7 .7 .7], [0 0 0], [.4 .4 .4], [.7 .7 .7], [0 0 0]};
    if cfg.grayscale, sty = gs; else sty = cs; end;
    sty.markers =   {'o',   's',    '^',    'v',    'd',    '*',    'X',    'h'};
    sty.markersizes = [4    4       4.5     4.5     4.5     5       4.5,    4.5];
    nsty = length(sty.colors); 
    % performance style
    sty.perf         = cell(nsty, 1);    % style for mean plot
    sty.perfedgearea = cell(nsty, 1);    % style for minmax/conf edge with shading
    sty.perfedgeline = cell(nsty, 1);    % style for minmax/conf edge without shading
    sty.patch        = cell(nsty, 1);    % shading style
    % convergence styles
    sty.conv         = cell(nsty, 1);    % mean
    for i = 1:length(sty.colors),
        sty.perf{i} = {'Marker', sty.markers{i}, 'MarkerSize', sty.markersizes(i), ...
            'Color', sty.colors{i}, 'MarkerFaceColor', sty.colors{i}};
        sty.perfedgearea{i} = {'Color', sty.colors{i}};
        sty.perfedgeline{i} = {'Color', sty.colors{i}}; % for now, the same
%         sty.perfedgeline{i} = {'Color', sty.colors{i}, ...
%             'Marker', sty.markers{i}, 'MarkerSize', sty.markersizes(i), 'MarkerFaceColor', sty.colors{i}};
        if cfg.transparency, % use alpha 
            sty.patch{i} = {sty.colors{i}, 'edgecolor' , 'none', 'facealpha', 0.3};
        else        % no alpha use a lighter gray than the color to shade the patch
            % scale the edge color into the interval [.5, .9] to obtain a
            % light patch color corresponding to the edge color
            % 2010-05-12: removed the "mean" to get colored patches
            patchcolor = .6 + rgb(sty.colors{i}) * (.9 - .6);
            % patchcolor = .6 + mean(rgb(sty.colors{i})) * (.9 - .6);
            sty.patch{i} = {[0 0 0] + patchcolor, 'edgecolor' , 'none'};
        end;
        sty.conv{i} = {'Color', sty.colors{i}};
    end;
    % common style for average, max, min performance
    sty.avg = {'-'}; sty.lower = {':'}; sty.upper = {'--'};
    sty.opt = {'-', 'Color', 'k', 'LineWidth', 1.5};
    
    % turn styles around for addtocrt (so that identical styles are less likely)
    if cfg.addtocrt,
        sty.colors = {sty.colors{end:-1:1}};
        sty.markers = {sty.markers{end:-1:1}};
        sty.markersizes = sty.markersizes(end:-1:1);
        sty.conv = {sty.conv{end:-1:1}};
        sty.perf = {sty.perf{end:-1:1}};
        sty.conv = {sty.conv{end:-1:1}};
        sty.patch = {sty.patch{end:-1:1}};
    end;
        
    figh = [];
    % init stuff depending on whether we're creating new plot or adding to current
    % note we don't support perf & conv plots at the same time
    if any(cfg.addtocrt),        % retrieve current figure handle and legend
        % check if handle to a figure (if yes add to that figure) or just boolean flag
        if ~islogical(cfg.addtocrt),    figh(end+1) = cfg.addtocrt(1); figure(figh(end));
        else                            figh(end+1) = gcf;
        end;
        [legh, objh, lh, leg] = legend; clear legh objh;
    else                    % create new figure
        figh(end+1) = figure; lh = []; leg = {};
    end;
    
    if cfg.onlineperf || cfg.onlinereturn,          % plot performance/return using PLOT, and PERF styles
        if cfg.onlineperf,          quant = 'J';
        elseif cfg.onlinereturn,    quant = 'ret';
        end;
        for iexp = 1:cfg.nexp,
            % shorthand vars
            quantvalues = STATS.(quant){iexp}; 
            tvalues = STATS.([quant 'time']){iexp}(1, :); 
            explabel = STATS.lab{iexp};
            hold on;
%             lh(end+1) = errorbar(tvalues, mean(quantvalues, 1), std(quantvalues), sty.avg{:}, sty.perf{iexp}{:});
            lh(end+1) = cfg.plotfun(tvalues, mean(quantvalues, 1), sty.avg{:}, sty.perf{iexp}{:});
            if cfg.repeat == 1 || (~cfg.plotminmax && ~cfg.plotconf),
                leg{end+1} = [cfg.expprefix ' ' explabel];
            elseif cfg.repeat > 1 && (cfg.plotminmax || cfg.plotconf),  % plot min and max bounds
                % acount for empty label to prettify comma location
                if isempty(explabel),
                    if isempty(cfg.expprefix),  leg{end+1} = lab.avgpref;
                    else                        leg{end+1} = [cfg.expprefix ', ' lab.avgpref];
                    end;
                else    leg{end+1} = [cfg.expprefix ' ' explabel ', ' lab.avgpref];
                end;
                if cfg.plotconf,
                    switch cfg.conftype,
                        case 'gauss',
                            lowconf = mean(quantvalues) - (1.96 .* std(quantvalues) ./ sqrt(size(quantvalues, 1)));
                            uppconf = mean(quantvalues) + (1.96 .* std(quantvalues) ./ sqrt(size(quantvalues, 1)));
                        case 'ttest',
                            [hdummy, pdummy, confint] = ttest(quantvalues);
                            lowconf = confint(1, :); uppconf = confint(2, :);
                    end;
                    if strcmp(cfg.minmaxtype, 'area'),
                        % place patch "under" the other graphics
                        patch([tvalues fliplr(tvalues)], [lowconf fliplr(uppconf)], ...
                            zeros(length(tvalues)*2)-.1, sty.patch{iexp}{:});
                        % note both bounds use the same style
                        lh(end+1) = cfg.plotfun(tvalues, lowconf, sty.upper{:}, sty.perfedgearea{iexp}{:});
                        cfg.plotfun(tvalues, uppconf, sty.upper{:}, sty.perfedgearea{iexp}{:});
                    else    % use "line" style for edges
                        % note both bounds use the same style
                        lh(end+1) = cfg.plotfun(tvalues, lowconf, sty.upper{:}, sty.perfedgeline{iexp}{:});
                        cfg.plotfun(tvalues, uppconf, sty.upper{:}, sty.perfedgeline{iexp}{:});
                    end;
                    leg{end+1} = '95% confidence bounds';
                elseif cfg.plotminmax,
                    if strcmp(cfg.minmaxtype, 'area'),
                        % place patch "under" the other graphics
                        patch([tvalues fliplr(tvalues)], [min(quantvalues) fliplr(max(quantvalues))], ...
                            zeros(length(tvalues)*2)-.1, sty.patch{iexp}{:});
                        lh(end+1) = cfg.plotfun(tvalues, min(quantvalues), sty.lower{:}, sty.perfedgearea{iexp}{:});
                        lh(end+1) = cfg.plotfun(tvalues, max(quantvalues), sty.upper{:}, sty.perfedgearea{iexp}{:});
                    else    % use "line" styles for edges                        
                        lh(end+1) = cfg.plotfun(tvalues, min(quantvalues), sty.lower{:}, sty.perfedgeline{iexp}{:});
                        lh(end+1) = cfg.plotfun(tvalues, max(quantvalues), sty.upper{:}, sty.perfedgeline{iexp}{:});
                    end;
                    % leg{end+1} = [cfg.expprefix ' ' explabel ' ' lab.minpref];
                    % leg{end+1} = [cfg.expprefix ' ' explabel ' ' lab.maxpref];
                    leg{end+1} = lab.minpref;
                    leg{end+1} = lab.maxpref;
                end;
            else
                warning('APPROXRL:warning', 'Incorrect value of cfg.repeat');
            end;
%             pause;
        end;
        ylabel(lab.(quant)); 
    elseif cfg.onlineconv,      % plot convergence using SEMILOGY, and CONV styles
        quant = 'deltah';
        for iexp = 1:cfg.nexp,
            % shorthand vars
            deltah = STATS.deltah{iexp}; deltahtime = STATS.deltahtime{iexp}(1, :); 
            explabel = STATS.lab{iexp};
            lh(end+1) = semilogy(deltahtime, mean(deltah, 1), sty.avg{:}, sty.conv{iexp}{:});
            hold on;
            leg{end+1} = [cfg.expprefix ' ' explabel];
%             if cfg.repeat > 1 && cfg.plotminmax,  % plot min and max bounds (unused)
%                 lh(end+1) = semilogy(deltahtime, min(deltah), sty.lower{:}, sty.conv{iexp}{:});
%                 lh(end+1) = semilogy(deltahtime, max(deltah), sty.upper{:}, sty.conv{iexp}{:});
%                 leg{end+1} = [cfg.expprefix ' ' explabel ' ' lab.minpref];
%                 leg{end+1} = [cfg.expprefix ' ' explabel ' ' lab.maxpref];
%             end;
        end;
        ylabel(lab.deltah{:}); 
    end;
    % set the legend, labels, etc.
    lh = legend(lh, leg{:}); set(lh, 'Location', 'Best'); % , 'FontSize', 8
    if ~any(cfg.addtocrt), xlabel(lab.time); end;
    box on;     % before setfigprop so that it can be overridden by setfigprop
    setfigprop(cfg); 
    % if addtocrt is explicitly set to 0, things will be added to the figure
    if isempty(cfg.addtocrt) || ~(cfg.addtocrt == 0),
        if cfg.fixedfigname,
            fname = [cfg.savedir cfg.savefig];
        else        
            if isempty(cfg.miscinfo),
                fname = [cfg.savedir cfg.savefig '_' lab.([quant 'file'])];
            else
                fname = [cfg.savedir cfg.savefig '_' cfg.miscinfo '_' lab.([quant 'file'])];
            end;
        end;
        saveplot(figh(end), fname, cfg.plottarget);
    end;
    varargout = {figh, STATS};
    % dirty trick to avoid passing arguments through various scripts
    LSPIEXP_STATS = STATS;
end;

% cleanup
if cfg.cleanup,
    for iexp = 1:cfg.nexp,
        par = get_params(iexp, cfg.flatparams(:, iexp), cfg.expanded, EXPANDABLE, cfg);
        for rep = cfg.startat:(cfg.startat + cfg.repeat - 1),
            df = gen_dfname(cfg, par, cfg.expanded, EXPANDABLE, rep);
            dispx(['Level-' num2str(cfg.cleanup) ' cleanup of [' df ']'], cfg.verb, 1);
            lspiexp_cleanup(df, cfg.cleanup);
        end;
    end;    % experiments loop
end;   

% make sure we're returning at least one argument, for compatibility
if ~exist('varargout', 'var'), varargout = {[]}; end;

end     % LSPIEXP =============================================



% --------------------
% Local functions

function [expanded, flatparams, varying] = expand_params(cfg, EXPANDABLE)
expanded = {'N', 'M', 'Ns'};    % these will always be expanded
paramvalues = {cfg.N, cfg.M, cfg.Ns};
varying = {};
if length(cfg.N) > 1, varying{end+1} = 'N'; end;
if length(cfg.M) > 1, varying{end+1} = 'M'; end;
if length(cfg.Ns) > 1, varying{end+1} = 'Ns'; end;
for i = 1:length(EXPANDABLE),
    if length(cfg.(EXPANDABLE{i})) > 1 && ~any(strcmp(EXPANDABLE{i}, cfg.donotexpand)) ...
            || (strcmp(EXPANDABLE{i}, 'hNs') && cfg.uselspih && cfg.adaptiveNs),
        % # of h approximator samples needs to vary if adaptive
        expanded{end+1} = EXPANDABLE{i};
        varying{end+1} = EXPANDABLE{i};
        pval = cfg.(EXPANDABLE{i});
        if isempty(pval),   paramvalues{end+1} = NaN;
        else                paramvalues{end+1} = pval;
        end;
    end;
end;
flatparams = flat(paramvalues);
% adaptive Ns following the # of params in the approximators
if cfg.adaptiveNs,
    % first, check that Ns doesn't already vary
    if length(cfg.Ns) > 1,
        error('Cannot use adaptive Ns when multiple values of Ns are given.');
    elseif cfg.uselspih && length(cfg.hNs) > 1,
        error('Cannot use adaptive hNs with LSPIH when multiple values of hNs are given.');
    end;
    % get the model order
    minfo = feval(cfg.problem, 'info');
    % compute the number of parameters in the approximators for every experiment
    nexp = size(flatparams, 2);
    Napp = zeros(1, nexp);
    if cfg.uselspih, Nhapp = zeros(1, nexp); end;
    for iexp = 1:size(flatparams, 2),
        par = get_params(iexp, flatparams(:, iexp), expanded, EXPANDABLE, cfg);
        switch cfg.type,
            case {'triang', 'rbfdisc'},
                Napp(iexp) = par.N^minfo.p * par.M^minfo.q;
            case {'triangpoly', 'rbfpoly', 'triangortho', 'rbfortho'},
                Napp(iexp) = par.N^minfo.p * (par.M+1);
            case {'rbfquad', 'triangquad'},
                Napp(iexp) = par.N^minfo.p * 3;
        end;
        if cfg.uselspih,
            switch cfg.htype,
                case {'rbfsing', 'triangsing'}
                    Nhapp(iexp) = par.N^minfo.p;
            end;
        end;
    end;
    % set the number of samples, taking the first experiment as the baseline
    flatparams(3, :) = ceil(Napp ./ Napp(1) .* cfg.Ns(1));
    if cfg.uselspih, 
        ihNs = find(strcmp('hNs', expanded), 1, 'first');   % find index of hNs parameter
        flatparams(ihNs, :) = ceil(Nhapp ./ Nhapp(1) .* cfg.hNs(1)); 
    end;
end;
end

function par = get_params(iexp, paramvector, expanded, EXPANDABLE, cfg)
for i = 1:length(expanded),
    par.(expanded{i}) = paramvector(i);
end;
% the rest are constant
for i = 1:length(EXPANDABLE),
    pname = EXPANDABLE{i};
    if ~any(strcmp(pname, expanded)),   
        if length(cfg.(pname)) <= 1, 
            % use single parameter from the config
            par.(pname) = cfg.(pname);
        else
            % assuming # of param values is identical to number of experiments
            % retrieve param value corresponding to this experiment
            par.(pname) = cfg.(pname)(iexp);
        end;
    end;
end;
end

function [sp, si] = find_params(p, names, sfp, sexpanded, EXPANDABLE, scfg)
% Looks for parameters 'sp' within 'sfp' that have the same values for fields 'names' as 'p'
for i = 1:size(sfp, 2),
    sp = get_params(sfp(:, i), sexpanded, EXPANDABLE, scfg);
    eq = 1; for j = 1:length(names), eq = eq && (p.(names{j}) == sp.(names{j})); end;
    if eq, si = i; break; end;
end;
if ~eq, sp = []; si = []; end;   % not found
end

% --------------------
function df = gen_dfname(cfg, par, expanded, EXPANDABLE, rep)     
% Subfunction to generate datafile
if ~isempty(cfg.fixeddf),   
    % use a fixed file name
    df = cfg.fixeddf;
else    % generate filename automatically
    % custom parameter names
    PARNAMES.N = 'N'; PARNAMES.M = 'M'; PARNAMES.Ns = 'Ns'; PARNAMES.hNs = 'Ns'; PARNAMES.hN = 'N';
    PARNAMES.gamma = 'g'; PARNAMES.explor = 'e'; 
    PARNAMES.dupd = 'dupd'; PARNAMES.trialtime = 'tt'; PARNAMES.explordecay = 'ed';
    PARNAMES.alpha = 'a';
    % uses algorithm and problem type to generate datafile prefix
    switch cfg.alg,
        case 'lspibatch',       df = 'lspib';
        case 'lspionline',      df = 'lspio';
        case 'lspihonline',     df = 'lspioh';
        case 'lspeonline',      df = 'lspeo';
        otherwise,              df = cfg.alg;
    end;
    switch cfg.problem
        case 'dc_problem',      df = [df 'dc'];
        case 'invpend_problem', df = [df 'ip'];
        case 'ipsetup_problem', df = [df 's'];
        case 'rarm_problem',	df = [df 'rarm'];
    end;
    % infix
    if ~isempty(cfg.infix),
        df = [df '_' cfg.infix];
    end;
    % approx
    if ~any(strcmp(cfg.hide, 'approx')),
        df = [df '_' cfg.type];
        if ~any(strcmp(cfg.hide, 'N')), 
            df = [df PARNAMES.N num2str(par.N) cfg.xspace];
        end;    % if hide N
        if ~any(strcmp(cfg.hide, 'M')) 
            switch cfg.type,
                case {'triang', 'rbfdisc', 'polydisc', 'rbflindisc'},
                    df = [df PARNAMES.M num2str(par.M) cfg.uspace];
                case {'triangpoly', 'rbfpoly', 'triangortho', 'rbfortho'},
                    df = [df PARNAMES.M num2str(par.M)];
                case {'rbfquad', 'triangquad'},
            end;
                    % nothing
        end;    % if hide M
    end;        % if hide approx
    % samples
    if ~cfg.useonline && ~any(strcmp(cfg.hide, 'samples')),
        df = [df '_' cfg.method];
        if ~any(strcmp(cfg.hide, 'Ns')),
            df = [df PARNAMES.Ns num2str(par.Ns)];
        end;    % if hide Ns
    end;        % if hide samples
    % ==== Policy approx
    if cfg.uselspih,
        % approx
        if ~any(strcmp(cfg.hide, 'happrox')),
            df = [df '_' cfg.htype];
            if ~any(strcmp(cfg.hide, 'hN')),
                df = [df PARNAMES.hN num2str(par.hN) cfg.hxspace];
            end;    % if hide N
        end;        % if hide approx
        % samples
        if ~any(strcmp(cfg.hide, 'samples')),
            df = [df '_' cfg.hmethod];
            if ~any(strcmp(cfg.hide, 'hNs')),
                df = [df PARNAMES.hNs num2str(par.hNs)];
            end;    % if hide Ns
        end;        % if hide samples
    end;
    processed = {'N', 'M', 'Ns', 'hN', 'hNs'};
    % modified 9-12-2008
    % always show the parameters in the order in which they appear in EXPANDABLE
    %   (regardless of whether they are expanded or not)
    for i = 1:length(EXPANDABLE),
        pname = EXPANDABLE{i};
        % disregard parameter if already processed
        if any(strcmp(pname, processed)), continue; end;
        if any(strcmp(pname, expanded)) ,
            % add, but only if not hidden
            if ~any(strcmp(pname, cfg.hide)),
                df = [df '_' PARNAMES.(pname) num2strx(par.(pname))];
            end;
        else
            % show if forced to show (otherwise, being non-expanded, it is not shown)
            if any(strcmp(pname, cfg.show)),
                df = [df '_' PARNAMES.(pname) num2strx(par.(pname))];
            end;
        end;
    end;
    
%     % NOTE earlier file names generated using the code below may no longer be found
%     % In that case, use compatibility code below
%     % show non-processed, expanded params that are not hidden
%     for i = 1:length(expanded),
%         if ~any(strcmp(expanded{i}, processed)) && ~any(strcmp(expanded{i}, cfg.hide)),
%             df = [df '_' PARNAMES.(expanded{i}) num2strx(par.(expanded{i}))];
%         end;
%     end;
%     % show non-processed, constant params that are forced to be shown
%     for i = 1:length(EXPANDABLE),
%         if any(strcmp(EXPANDABLE{i}, cfg.show)) && ...
%                 ~any(strcmp(EXPANDABLE{i}, processed)) && ~any(strcmp(EXPANDABLE{i}, expanded)),
%             df = [df '_' PARNAMES.(EXPANDABLE{i}) num2strx(par.(EXPANDABLE{i}))];
%         end;
%     end;
end;
% postfix
if ~isempty(cfg.postfix),
    df = [df '_' cfg.postfix];
end;
% repetition index, only if there's more than 1 rep to run, or forced show
if ~any(strcmp(cfg.hide, 'rep')) && (cfg.repeat > 1 || any(strcmp(cfg.show, 'rep'))),
    df = [df '_#' num2str(rep)];
end;
end     % function GEN_DFNAME


% --------------------
% Subfunction to cleanup an LSPI datafile, thus optimizing on the HDD space taken
function lspiexp_cleanup(cleanupdf, level)
% Levels:
%   0.5 	soft cleanup, will compute conditioning numbers for A if they were not computed
%       while running the algorithm. All the new data files should have condition numbers saved,
%       so that this mode should no longer be necessary
%   1       normal cleanup, history of A, b, the values of A, b, and Ainv. Online experiments
%   can no longer be resumed
%   1.5     medium cleanup, cleans up PHI and HPHI (precomputed BF values)
%       These take a lot of space and not so much time to recompute.
%   2       hard cleanup, cleans up PHI and HPHI, histories of parameter vectors. This means
%       the evolution of the solution over iterations can no longer be examined.
%       Should only be used as a last resort, as parameter vectors don't really take up that
%       much space

% ??? TODO is this suitable for online experiments?

if ~exist([cleanupdf '.mat'], 'file'), return; end;
% make sure the same cleanupdf is kept
cdf2 = cleanupdf; load(cleanupdf); cleanupdf = cdf2; clear cdf2;
if exist('cleanedup', 'var') && (cleanedup >= level), return; end;  % nothing to do 

% remove thetastar in any case, it is no longer used
clear thetastar

K = find(~isnan(deltah), 1, 'last') - 1;    % may be needed by various cleaning levels

% soft cleanup: compute conditioning numbers for A matrices (unless already computed) and store them
if level == 0.5,
    if ~exist('condA', 'var'),
        condA = zeros(1, K);
        for k=1:K,
            if issparse(Ah{k}),     condA(k) = condest(Ah{k});
            else                    condA(k) = cond(Ah{k});
            end;
        end;            
    end;
end;

% this happens in any cleanup procedure
clear Ah Ainvh bh Ainv A B blk
% these only have effect for lspionline with recinv
clear col deltaAinv den uvT

if level >= 1.5,
    % semi-hard cleanup: cleanup A0, b, PHI but keep parameter histories
    % experiment can be resumed but some computation will have to be re-done
    clear PHI HPHI A0 b OFF
    if level >= 2,      
        % hard cleanup: no evolution replay possible; no resume possible if relying on precomputed PHI
        clear thetah hthetah
    end;
end;
cleanedup = level;
clear level;
save(cleanupdf);

end


% Obsolete code

% function pv = params2vector(par, expanded)
% % transform params into vector
% pv = zeros(length(expanded), 1);
% for i = 1:length(expanded), pv(i) = par.(expanded{i}); end;
% end
% function par = vector2params(pv, expanded)
% % transform params into vector
% par = struct;
% for i = 1:length(expanded), par.(expanded{i}) = pv(i); end;
% end

%                 if isempty(fe),
%                     dispx('Did not find exact sample match. Trying only Ns.', cfg.verb, 4);
%                     fe = find(sfp(3, :) == par.Ns, 1, 'first');
%                     if isempty(fe),
%                         error('Did not find an Ns-match in UseSamples.');
%                     else
%                         spar = get_params(sfp(:, fe), sexpanded, EXPANDABLE, scfg);
%                         ecfg.loadsamples = gen_dfname(scfg, spar, sexpanded, EXPANDABLE, rep);
%                         dispx(['Using Ns-only samples match [' ecfg.loadsamples ']'], cfg.verb, 4);
%                     end;
%                 else
