function  [uroot, planinfo, tree] = opd_structures(xroot, model, cfg)
% OPD Optimistic planning for deterministic systems.
%   Detailed explanation goes here
% Implemented using structures.
% Author: Lucian Busoniu
% Versions:

CFG.gamma = [];             % discount factor
CFG.n = [];                 % expansion budget
CFG.criterion = 'optim';    % expansion criterion: optim(istic) or unif(orm)
CFG.U = [];                 % discrete, flat action space
% display config
CFG.verb = 0;               % verbosity: the higher, the more detailed the messages displayed

cfg = checkparams(cfg, CFG);

% quick-access to config parameters
% (note to avoid conflict with 'gamma' function, we must name discount factor differently)
discount = cfg.gamma;
structtovars(cfg, {'n', 'U', 'criterion'});
M = size(U, 2);      % number of discrete actions

% start counting time
t = cputime;

% prepare empty node template
nodetpl = struct('x', [], 'term', [], 'util', [], ...
                'leaf', [], 'depth', [], ...
                'parent', [], ...
                'children', [], 'r', []);

% initialize the tree to contain a single root, leaf node corresponding to xroot
tree = nodetpl;
tree.x = xroot;
tree.term = false;   % presumably, we are not asked to make a decision in a terminal state
tree.util = 0;
% meta-info
tree.leaf = true;
tree.depth = 0;

% main tree construction loop
for tau = 1:n,
    
    % 1) find all the leaves
    I = find([tree(:).leaf]);
    
    % 2) select a leaf node to expand
	% TODO is there a problem with the terminal states here? if max-b-value state is terminal
	% algorithm has nothing to do
    if criterion(1) == 'o',
        % optimistic: maximal b-value
        B = [tree(I).util] + ~[tree(I).term] .* (discount .^ [tree(I).depth]) ./ (1-discount);
        I = I(B == max(B));
    elseif criterion(1) == 'u',
        % uniform: minimal depth
        I = I([tree(I).depth] == min([tree(I).depth]));
    else error('Unknown criterion [%s]', criterion);
    end;
    % if all selected nodes are terminal, we are done
    if all([tree(I).term]), break; end;
    % else choose one deterministically and continue
    iexp = I(1);
    
    % 3) expand the chosen node
    tree(iexp).leaf = false;
    % init children and transition information tables
    tree(iexp).children = nan(M, 1);
    tree(iexp).r = nan(M, 1);
    % run through the discrete actions 
    for j = 1:M,
        [x, r, term] = feval(model.fun, model, tree(iexp).x, U(:, j));
        if r < 0 || r > 1, error('Rewards must be in [0, 1]'); end;
        tree(iexp).r(j) = r;

        % find index where new node will be, save it on the children table of current node
        inew = length(tree) + 1;
        tree(iexp).children(j) = inew;

        % create new node
        tree(inew) = nodetpl;
        % data
        tree(inew).x = x;
        tree(inew).term = term;
        % compound partial return, starting from parent
        tree(inew).util = tree(iexp).util + discount^tree(iexp).depth * r;
        % meta-info
        tree(inew).parent = iexp;                   % parent is expanded node
        tree(inew).leaf = true;                     % this is a leaf
        tree(inew).depth = tree(iexp).depth + 1;    % one deeper than the parent
    end;
    
    % 4) update utilities affected by expanded node
    i = iexp;
    while i >= 1,       % while root or above
        tree(i).util = max([tree(tree(i).children).util]);
        % move to parent (or 0=stop if we are at root)
        if i > 1,   i = tree(i).parent;
        else        i = 0;
        end;
    end;
    
    if cfg.verb >= 2, visualizetree(tree, U, discount, iexp); pause; end;
    
end;    % end main tree construction loop

if cfg.verb >= 2, 
    disp(sprintf('Tree from x=%s done: n=%d expansions, depth=%d', ...
        num2strx(xroot, 'vector'), tau, max([tree(:).depth]))); 
end;

% finally, return a greedy action at the tree root
cutil = [tree([tree(1).children]).util];
j = find(cutil == max(cutil), 1, 'first');
uroot = U(:, j);

if nargout >= 2,
    planinfo.n = tau;
    planinfo.depth = max([tree(:).depth]);
    planinfo.trun = cputime - t;
    planinfo.message = sprintf('Planning done, n=%d expansions, depth=%d', tau, planinfo.depth); 
end;

end % main function



% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Local function to plot a graphical representation of the current tree
function visualizetree(tree, U, discount, iexp)
persistent figh;
figsize = [800,600];
if isempty(figh), figh = figurex(struct('size', figsize)); else figure(figh); end;
clf;

% initialize helper variables for nested functions
M = size(U, 2);
nodestoexpand = {};
% build and customize tree
guitree = uitree('Root', rendernode(1));
guitree.Position(3) = figsize(1);
% expand nodes
for i = 1:length(nodestoexpand), guitree.expand(nodestoexpand{i}); end;
% done

    % %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Nested function to render a node recursively
    function guinode = rendernode(i, jp, rp)
    % create label
    if i == 1,  label = 'root: ';
    else        label = ['u=' num2str(U(:, jp)') ' (r=' num2str(rp) ')   #-->   '];
    end;
    label = [label 'x=' num2str(tree(i).x')];
    if tree(i).term, label = [label ' (T)']; end;
    if tree(i).leaf, 
        label = [label '; b=' num2str(tree(i).util + ~tree(i).term * discount^tree(i).depth/(1-discount))]; 
    end;
    label = [label '; util=' num2str(tree(i).util)];
    if i == iexp, label = ['<html><font color="red">' label '</font></html>']; end;
    % add tree(i) to tree
    guinode = uitreenode(label, label, [], tree(i).leaf);
    if ~tree(i).leaf,
        % remember we need to expand this node
        nodestoexpand{end+1} = guinode;
        for j = 1:M,
            guinode.add(rendernode(tree(i).children(j), j, tree(i).r(j))); 
        end;
    end;
    end     % function rendernode
end     % function visualizetree

