function  [uroot, planinfo, tree] = opss_structures(xroot, model, cfg)
% OPSS Optimistic planning for sparsely stochastic systems
%   Old, compatibility variant, using nested structures

% Author: Lucian Busoniu
% Versions:
%   1.0     - unif random tie breaking, probability selection criterion
%   2.0     - renamed to opss, added subtree reuse
% Note on tree reuse: must rebuild the subtree due to the way pointers are handled. This helps as
% probabilities and returns are recomputed from scratch rather than adjusted, which would 
% increase numerical errors due to repeated addition/subtraction, multiplication/division.

CFG.gamma = [];             % discount factor
CFG.n = [];                 % expansion budget
CFG.criterion = 'contrib';  % expansion criterion: based on the whole tree: 'unif'
                            % based on optimistic subtree: 'contrib', 'return', 'smallest' (or 'shallowest'),
                            % 'xcontrib' (maximal contribution subject to constraints)
CFG.actsel = 'util';        % action selection at root: 'util', 'b'
CFG.L = [];                 % maximum number of successors for each state-action pair
CFG.U = [];                 % discrete, flat action space
CFG.reusetree = 0;          % try reusing previous tree (only when k is set and > 1)
CFG.maxtreesize = 50000;    % do not reuse tree if larger (has more nodes) than this
CFG.k = -1;                 % discrete time step
% display config
CFG.verb = 0;               % verbosity: the higher, the more detailed the messages displayed

persistent oldtree;

cfg = checkparams(cfg, CFG);

% quick-access to config parameters
% (note to avoid conflict with 'gamma' function, we must name discount factor differently)
discount = cfg.gamma;
structtovars(cfg, {'n', 'criterion', 'L', 'U'});
M = size(U, 2);      % number of discrete actions

% start counting time
t = cputime;

% prepare empty node template
nodetpl = struct('x', [], 'term', [], 'b', [], 'util', [], 'ret', [], 'prob', [], ...
                'leaf', [], 'depth', [], ...
                'parent', [], ...
                'children', [], 'f', [], 'r', [], ...
                'flagsmallprob', 0, 'flagintermdepth', 0);

tree = [];
if cfg.reusetree && ~isempty(oldtree) && cfg.k > 1,  % try reusing subtree (not at trajectory start though)
    % find all occurences of root state
    oldiroot = findflat(xroot, [oldtree(:).x]);
    if ~isempty(oldiroot),
         % choose the smallest-depth occurence
        [dmin, imin] = min([oldtree(oldiroot).depth]); oldiroot = oldiroot(imin);
        % copy root and set root-relative info
        tree = oldtree(oldiroot);
        tree.parent = []; tree.prob = 1; tree.ret = 0; tree.depth = 0;
        if tree.leaf, I = [];           % nothing to do
        else I = 1; oldI = oldiroot;     % init tree fronts with the root/root of interesting subtree
        end;
        % iterate until old subtree is exhausted
        while ~isempty(I),
            % these will hold the new front
            II = []; oldII = [];
            % for each node on the current front
            for ii = 1:length(oldI),
                oldi = oldI(ii); i = I(ii);
                % init children table
                tree(i).children = nan(M, L);
                % get all children from old tree, updating pointers and root-relative info
                for j = 1:M,
                    Lj = find(~isnan(oldtree(oldi).children(j, :)), 1, 'last');
                    for ell = 1:Lj,
                        oldichild = oldtree(oldi).children(j, ell);
                        ichild = length(tree)+1;
                        % copy child
                        tree(ichild) = oldtree(oldichild);
                        % update pointers
                        tree(i).children(j, ell) = ichild;
                        tree(ichild).parent = [i j];
                        % update root-relative info
                        tree(ichild).depth = tree(i).depth + 1;
                        tree(ichild).prob = tree(i).prob * tree(i).f(j, ell);
                        tree(ichild).ret = tree(i).ret + discount^tree(i).depth * tree(i).r(j, ell);
                        if ~tree(ichild).leaf, II(end+1) = ichild; oldII(end+1) = oldichild; end;
                    end;
                end;
            end;
            I = II; oldI = oldII;
        end;
    end;
end;
if isempty(tree),
    % initialize the tree to contain a single root, leaf node corresponding to xroot
    tree = nodetpl;
    tree.x = xroot;
    tree.term = false;   % presumably, we are not asked to make a decision in a terminal state
    tree.b = repmat(1/(1-discount), M, 1);
    tree.util = zeros(M, 1);
    tree.prob = 1;
    tree.ret = 0;
    % meta-info
    tree.leaf = true;
    tree.depth = 0;
end;

% visualize initial tree
if cfg.verb >= 2, visualizetree(tree, U); pause; end;

% main tree construction loop
for tau = 1:n,
    
    % 1) find the candidate leaves for expansion
    if criterion(1) == 'u',     % uniform expansion -- just get all the leaves
        I = find([tree(:).leaf]);
    else
        % greedy leaves (forward traversal)
        I = 1;  % init with root
        while ~all([tree(I).leaf]),         % not all leaves
            Inew = [];                      % new nodes to add
            for i = I(~[tree(I).leaf]),     % loop through each inner node
                % find a greedy branch, breaking ties deterministically (i.e. choose first greedy branch)
                [bmax jmax] = max(tree(i).b);
                % add children on greedy branch, accounting for the possibility of fewer than L children
                Ljmax = find(~isnan(tree(i).children(jmax, :)), 1, 'last');
                Inew = [Inew tree(i).children(jmax, 1:Ljmax)];
            end;
            % new set of nodes is old set of LEAF nodes, plus the newly added nodes irrespective of type
            I = [I([tree(I).leaf]) Inew];
        end;
    end;
    
    % 2) select a leaf node to expand
    % we know all nodes are leaves, but some of them may be terminal
    % if ALL are terminal, then we are sure of the action -- so we are done
    if all([tree(I).term]), break; end;
    % else, eliminate the terminal nodes and continue with the rest
    I = I(~[tree(I).term]);
    if criterion(1) == 'c',
        % largest contribution to the bound (breaking ties determ)
        % note we do not consider the factor 1/(1-gamma) as it appears everywhere
        % this is the variant used in the ADPRL paper
        C = [tree(I).prob] .* discount .^ [tree(I).depth];
        [vmax, imax] = max(C);
        iexp = I(imax);
    elseif criterion(1) == 's',
        % smallest depth - ensures gamma^d(iexp)/(1-gamma) > diam(hopt)
        % note this rule is similar to uniform expansion -- just the set of considered
        % candidate leaves differs -- here it's only the optimistic leaves
        [dmin imin] = min([tree(I).depth]);
        iexp = I(imin);
    elseif criterion(1) == 'x',
        % largest contribution; subject to the constraint gamma^d(iexp)/(1-gamma) > diam(hopt)
        % note we do not consider the factor 1/(1-gamma) as it appears everywhere
        B = discount .^ [tree(I).depth]; P = [tree(I).prob]; C = P .* B;
        % find leaves for which gamma^d(iexp)/(1-gamma) > diam(hopt)
        nearopt = B + 1e-12 >= sum(C);
        % find leaves with large probability
        % largeprob = P >= (1 ./ L.^[tree(I).depth]);
        % choose a leaf (ignoring the probabilities for now)
        [vmax, imax] = max(C .* nearopt);
        if vmax == 0, error('No near-optimal leaf - should not happen'); end;
        iexp = I(imax);
        tree(iexp).flagsmallprob = tree(iexp).prob < (1 / L^tree(iexp).depth);
        tree(iexp).flagintermdepth = tree(iexp).depth > min([tree(I).depth]);
        % visualizetree(tree, U, I, iexp, 1);
    elseif criterion(1) == 'r',     
        % largest upper bound weighted by probability (breaking ties determ)
        % this criterion is obsolete and should probably not be used
        P = [tree(I).prob];
        B = [tree(I).ret] + discount .^ [tree(I).depth] ./ (1-discount);
        [vmax, imax] = max(P .* B);
        iexp = I(imax);
    elseif criterion(1) == 'u',     
        % uniform: expand a node with the smallest depth (breaking ties determ)
        [dmin imin] = min([tree(I).depth]);
        iexp = I(imin);
    else error('Unknown criterion [%s]', criterion);
    end;
    
    % 3) expand the chosen node
    tree(iexp).leaf = false;
    % init children and transition information tables
    tree(iexp).children = nan(M, L);
    tree(iexp).f = nan(M, L);
    tree(iexp).r = nan(M, L);
    % run through the discrete actions 
    for j = 1:M,
        % use model to determine distribution over next states (+ rewards, terminal flags)
        [x, r, term, f] = feval(model.fun, model, tree(iexp).x, U(:, j));
        if any(r < 0) || any(r > 1), error('Rewards must be in [0, 1]'); end;
        % note we allow FEWER than L actions here
        if length(f) > L, error('Too many successors (%d, at most %d allowed)', length(f), L); end;
        % run through the possible next states
        for ell = 1:length(f),
            % save the transition probability and reward to tables of current node
            tree(iexp).f(j, ell) = f(ell);
            tree(iexp).r(j, ell) = r(ell);
            
            % find index where new node will be, save it on the children table of current node
            inew = length(tree) + 1;
            tree(iexp).children(j, ell) = inew;
            
            % create new node
            tree(inew) = nodetpl;
            % data
            tree(inew).x = x(:, ell);
            tree(inew).term = term(ell);
            if term(ell),   tree(inew).b = zeros(M, 1);
            else            tree(inew).b = repmat(1/(1-discount), M, 1);
            end;
            tree(inew).util = zeros(M, 1);
            % compound probability and partial return, starting from parent
            tree(inew).prob = tree(iexp).prob * f(ell);
            tree(inew).ret = tree(iexp).ret + discount^tree(iexp).depth * r(ell);
            % meta-info
            tree(inew).parent = [iexp j];               % parent is expanded node, along jth action
            tree(inew).leaf = true;                     % this is a leaf
            tree(inew).depth = tree(iexp).depth + 1;    % one deeper than the parent
        end;
    end;
    
    % 4) update b-values affected by expanded node
    % these are the b-values of the expanded node itself for all the actions, 
    % and those on the state-ACTION path from the expanded added node to the root
    i = iexp; jrange = 1:M;
    while i >= 1,       % while root or above
        for j = jrange,
            % account for the possibility of fewer successor states than the maximum L
            Lj = find(~isnan(tree(i).children(j, :)), 1, 'last');
            ichildren = tree(i).children(j, 1:Lj);
            tree(i).b(j) = tree(i).f(j, 1:Lj) * ...
                ( tree(i).r(j, 1:Lj) + discount * max([tree(ichildren).b]) )';
            % ichildren is a vector, [tree(ichildren).b] is a matrix with b-values of actions on
            % columns, so the max takes the maximum across actions and returns an L-length vector
            % that can be operated on in dimension sync with the rest
            tree(i).util(j) = tree(i).f(j, 1:Lj) * ...
                ( tree(i).r(j, 1:Lj) + discount * max([tree(ichildren).util]) )';
        end;
        if i > 1,       
            jrange = tree(i).parent(2);
            i = tree(i).parent(1);
        else    % i=1 => root, has no parent, stop the loop
            i = 0;
        end;
    end;
    
    if cfg.verb >= 2, visualizetree(tree, U, I, iexp); pause; end;
    
end;    % end main tree construction loop

% finally, return a greedy action at the tree root
% (greedy w.r.t. field actsel, which can be 'b' or 'util')
[vmax, jmax] = max(tree(1).(cfg.actsel));
uroot = U(:, jmax);

% save old tree if configured, and tree doesn't exceed maximum size
if cfg.reusetree, 
    if length(tree) <= cfg.maxtreesize, oldtree = tree; 
    else oldtree = [];
    end;
end;

% output stats if requested
if nargout >= 2,
    planinfo.n = tau;
    planinfo.depth = max([tree(:).depth]);
    planinfo.trun = cputime - t;
    planinfo.treesize = length(tree);
    planinfo.message = sprintf('Planning done, n=%d expansions, depth=%d, treesize=%d', ...
        tau, planinfo.depth, planinfo.treesize); 
end;

end % main function



% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Local function to plot a graphical representation of the current tree
function visualizetree(tree, U, I, iexp, optimonly)
if nargin < 3, I = []; end;
if nargin < 4, iexp = []; end;
if nargin < 5, optimonly = 0; end;
persistent figh;
figsize = [800,600];
if isempty(figh), figh = figurex(struct('size', figsize)); else figure(figh); end;
clf;

% initialize helper variables for nested functions
M = size(U, 2);
nodestoexpand = {};
% build and customize tree
guitree = uitree('Root', rendernode(1));
guitree.Position(3) = figsize(1);   % expand to figure width
% expand nodes
for i = 1:length(nodestoexpand), guitree.expand(nodestoexpand{i}); end;
% done

    % %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Nested function to render a node recursively
    function guinode = rendernode(i, jp, fp, rp)
    % create label
    if i == 1,  label = 'root: ';
    else        label = ['u=' num2str(U(:, jp)') ' (f=' num2str(fp) ', r=' num2str(rp) ')   #-->   '];
    end;
    label = [label 'x=' num2str(tree(i).x')];
    if tree(i).term, label = [label ' (T)']; end;
    label = [label '; b=' num2strx(tree(i).b, 'vector') '; prob=' num2str(tree(i).prob) ', ret=' num2str(tree(i).ret)];
    if i == iexp,       label = ['<html><font color="red">' label '</font></html>'];    % expanded
    elseif any(i == I), label = ['<html><font color="blue">' label '</font></html>'];   % on greedy subtree
    end;
    % add tree(i) to tree
    guinode = uitreenode(label, label, [], tree(i).leaf);
    if ~tree(i).leaf,
        % remember we need to expand this node
        nodestoexpand{end+1} = guinode;
        if optimonly,   [bopt, jrange] = max(tree(i).b); 
        else            jrange = 1:M;
        end;
        for j = jrange,
            Lj = find(~isnan(tree(i).children(j, :)), 1, 'last');
            for ii = 1:Lj,
                ic = tree(i).children(j, ii);
                guinode.add(rendernode(ic, j, tree(i).f(j, ii), tree(i).r(j, ii))); 
            end;
        end;
    end;
    end     % function rendernode
end     % function visualizetree


%         % largest contribution; subject to the constraint gamma^d(iexp)/(1-gamma) > diam(hopt)
%         % note we do not consider the factor 1/(1-gamma) as it appears everywhere
%         B = discount .^ [tree(I).depth];
%         P = [tree(I).prob];
%         C = P .* B;
%         % eliminate leaves not satisfiyng gamma^d(iexp)/(1-gamma) > diam(hopt)
%         C(B + 1e-12 < sum(C)) = -Inf;
%         % eliminate leaves not satisfying p(x) > 1/N^d(x)
%         C(P < 0.5 ./ L.^[tree(I).depth]) = -Inf;
%         % pick the maximum from the remaining
%         [vmax, imax] = max(C);
%         if vmax == -Inf,
%             error('No feasible leaf'); 
%         end;