function [figh, varargout] = qiresstudy(cfg)
% Study of Q-iteration as the number of (equidistant) fuzzy centers on each axis increases
% Assumes hyperbox Euclidean state space with bounds in model.maxx
% Supersedes fuzzyqi_resstudy
% Currently only supports fuzzy Qiteration

if nargin < 1,
    cfg = struct;
end;

% main config
CFG.run = 0;
CFG.overwrite = 0;
CFG.perf = 0;
CFG.runtime = 0;
CFG.iter = 0;
CFG.N = 3:11;           % numbers of fuzzy cores on each axis to try out
CFG.M = 3;              % numbers of discrete actions
CFG.problem = 'dc_problem';
% data file config
CFG.datafile = 'score'; % for saving the score
CFG.datadir = '';
CFG.dfprefix = [];
CFG.dfinfix = '';
% compatibility mode with fuzzyqi_resstudy: 
%   generate filenames of the form 'prefix_<Nvalue-1>x<Nvalue-1>_fuzzy' instead of
%   'prefix_N<Nvalue>M<Mvalue>'; reason for Nvalue-1 is that
%   fuzzyqi_resstudy generated N+1-length grids for resolution N
CFG.dfcompatibility = 0;
CFG.usedatafile = [];   % try getting already computed data from this file
% fuzzy QI config
CFG.gamma = 0.95;
CFG.maxiter = 1000;
CFG.eps = 1e-5;
CFG.term = 'ignore';
CFG.initstepdisp = .25;
CFG.itersave = 100;
CFG.silent = 1;
CFG.model_params = {};
% these fields get copied to the Qiteration config
QIFIELDS = {'problem', 'gamma', 'maxiter', 'eps', 'term', 'initstepdisp', 'itersave', 'silent', 'model_params', 'datadir'};
% score eval config
CFG.avgM = 0;
CFG.Jstar = [];
CFG.X0 = 'fine';
CFG.mc_eps = 0.01;
CFG.userew = '';

MCFIELDS = {'mc_eps', 'gamma', 'userew'};

% output config
CFG.verb = 2;
% plot config
CFG.legendprefix = ''; CFG.legendpostfix = '';
CFG.lsty = '-';  CFG.markertype = 'o';
CFG.grayscale = 1;
CFG.plotfun = @plot;        % to use for plotting
CFG.plot3fun = @mesh;       % to use for 3-D plotting
CFG = setfigprop(CFG, 'addfields');  % add figure settings fields to CFG
CFG.axislabelcase = 'sentence'; % default axis label case 
CFG.xlabel = [];        % any custom x label
CFG.addtocrt = 0;
CFG.eachaxis = 1;   % do not report total number of basis functions, but only on 1 axis

% % save configuration -- should be handled by setfigprop
% CFG.plottarget = 'screen';     % 'latex', 'beamer', 'screen', ''
% % only last created figure can be saved using these options
% CFG.savedir = ''; % 'D:\Work\tex\papers\alamas07\img\';
%                         % path for saving figure
% CFG.savefig = 'fzqires';    % filename for saving figure

% process config
if ischar(cfg), cfg = str2cfg(cfg, fieldnames(CFG)); end;
cfg = checkparams(cfg, CFG);
cfg.grayscale = grayscalefromconfig(cfg);
cfg.grayscale
if ~iscell(cfg.model_params),
    cfg.model_params = {cfg.model_params};
end;
cfg.NM = flat({cfg.N, cfg.M});
if isempty(cfg.usedatafile),
    cfg.usedatafile = cfg.datafile;
end;

cfg.nexp = size(cfg.NM, 2); % number of experiments

% echo
dispx(cfg, cfg.verb, 1);

model = feval(cfg.problem, 'model', cfg.model_params{:});

% run experiments
if cfg.run,
    for i = 1:cfg.nexp,
        N = cfg.NM(1, i); M = cfg.NM(2, i);
        % form grids
        xgrids = cell(model.p, 1);
        for p = 1:model.p,
            xgrids{p} = -model.maxx(p):(2*model.maxx(p)/(N-1)):model.maxx(p);
        end;
        ugrids = cell(model.q, 1);
        for q = 1:model.q,
            ugrids{q} = -model.maxu(q):(2*model.maxu(q)/(M-1)):model.maxu(q);
        end;
        
        % form filename
        df = gen_dfname(cfg, N, M);

        % create QI config and run QI if needed
        if ~exist([df '.mat'], 'file') || cfg.overwrite,
            dispx(['Running Q-iteration, N=' num2str(N) ', M=' num2str(M)], cfg.verb, 2);
            qicfg = struct;
            qicfg.run = 1;
            qicfg.xgrids = xgrids;
            qicfg.ugrids = ugrids;
            qicfg.datafile = df;
            qicfg = copyfields(cfg, qicfg, QIFIELDS);
            fuzzyqi(qicfg);
        end;
    end;
end;

% load data
if cfg.perf || cfg.iter || cfg.runtime,
    DATAVARS = {'data', 'cfg', 'X0'};
    % get X0 & make sure it is in the "cell array of grids" format
    % note it is not actually necessary to get the numerical X0 here since mc_fuzzyqi would
    % handle that; however we keep this behavior so that any cached data 
    % saved with the same X0 satisfies the X0 tests
    X0 = feval(cfg.problem, 'X0', cfg.X0); 
    if ~iscell(X0), X0 = unflat(X0); end;
    N0 = [length(X0{1}), length(X0{2})];
    data.J = cell(cfg.nexp, 1); data.eJ = nan(cfg.nexp, 1); 
    data.runtime = nan(cfg.nexp, 1); data.iter = nan(cfg.nexp, 1);
    
    % try getting any data that was already computed
    found = false(1, cfg.nexp);
    if exist([cfg.usedatafile '.mat'], 'file'),
        loadwithprefix(cfg.usedatafile, 'd_', DATAVARS);
        % data is no good unless X0 and MC eps were the same
%         (d_cfg.mc_eps == cfg.mc_eps) 
%         isequal(d_cfg.userew, cfg.userew) 
%         isequal(d_X0, X0), 
        if (d_cfg.mc_eps == cfg.mc_eps) ...
                && isequal(d_cfg.userew, cfg.userew) && isequal(d_X0, X0), 
            for i = 1:cfg.nexp,
                % data is good if same N and M, and MC score was computed
                di = d_cfg.NM == repmat(cfg.NM(:, i), 1, d_cfg.nexp);
                di = find(all(di), 1);
                if ~isempty(di) && ~isempty(d_data.J{di}),
                    found(i) = 1;
                    data.J{i} = d_data.J{di};
                    data.eJ(i) = d_data.eJ(di);
                    data.runtime(i) = d_data.runtime(di);
                    data.iter(i) = d_data.iter(di);
                end;
            end;
            dispx(['Found ' num2str(sum(found)) ' pre-computed scores out of ' num2str(cfg.nexp)], cfg.verb, 1);
        end;
    end;
    for i = find(~found),
        N = cfg.NM(1, i); M = cfg.NM(2, i);
        df = gen_dfname(cfg, N, M);
        if ~exist([df '.mat'], 'file'),
            error(['File [' df '] not found. Aborting.']);
        end;
        dispx(['Evaluating score, N=' num2str(N) ', M=' num2str(M)], cfg.verb, 2);
        if cfg.perf,    % only perform monte-carlo if perf plot is requested
            mccfg = struct; 
            mccfg.datafile = df;
            mccfg.X0 = X0;
            mccfg = copyfields(cfg, mccfg, MCFIELDS);
            [eJ, J] = mc_fuzzyqi(mccfg);
            data.eJ(i) = eJ; data.J{i} = J;
%             df
%             eJ
%             save eJdatatmp data 
        end;
        load(df, 'qistats');
        data.runtime(i) = qistats.t + qistats.tinit;
        % #iter computed from the delta record
        data.iter(i) = length(qistats.delta);
    end;
    if any(~found),
        if ~isempty(cfg.datadir) && (cfg.datadir(end) == '\'), cfg.datadir = cfg.datadir(1:end-1); end;
        save([cfg.datadir '\' cfg.datafile], DATAVARS{:});
    end;
end;

% ----------------------
% Plot code

% grayscale styles
if ischar(cfg.lsty),
    gs.perfstyle = {cfg.lsty, 'Color', 'k', 'Marker', cfg.markertype, 'MarkerSize', 4};
elseif iscell(cfg.lsty),
    gs.perfstyle = cfg.lsty;
end;
gs.perf3style = {'EdgeColor', 'k', 'Marker', cfg.markertype, 'MarkerSize', 2.25, 'MarkerFaceColor', 'k'};
% gs.perf3style = {'EdgeColor', 'k', 'Marker', cfg.markertype, 'MarkerSize', 2.5, 'MarkerFaceColor', 'k'};
gs.runtimestyle = gs.perfstyle; gs.iterstyle = gs.perfstyle;
gs.optstyle = {'-', 'Color', 'k', 'LineWidth', 2};
gs.cm = gray;
% color styles, with the appropriate changes from grayscale style
cs = gs;        
cs.cm = jet;
if ischar(cfg.lsty),
    cs.perfstyle = {cfg.lsty, 'Color', [0 .6 0], 'Marker', cfg.markertype, 'MarkerSize', 3, 'LineWidth', 1.3};
end;
cs.runtimestyle = cs.perfstyle; cs.iterstyle = cs.perfstyle;
cs.perf3style = {'Marker', cfg.markertype, 'MarkerSize', 4, 'MarkerFaceColor', 'flat'};
% set style
if cfg.grayscale, sty = gs; 
else sty = cs; end;

% readable labels
% commonprop = {'Interpreter', 'LaTeX', 'FontSize', 13}; % 2009-10-28: removed LaTeX labels
commonprop = {'Interpreter', 'TeX'};
if ~isempty(cfg.xlabel),        % allow a custom X label
    rl.x = cfg.xlabel;
else
    if cfg.eachaxis,
%         rl.x = 'Number of MFs on each axis, N''';
    %     rl.x = 'Number of BFs on each axis';
        rl.x = 'N''';
    else
%         rl.x = 'Number of fuzzy sets, N';
    %     rl.x = 'Number of BFs';
        rl.x = 'N';
    end;
end;
% rl.u = 'Number of discrete actions, M';
rl.u = 'M';
% labels to be used in axis titles
rl.perf = 'Score'; rl.runtime = 'Execution time [s]'; rl.iter = 'Iterations';
% labels to be used in legends
rl.perflegend = 'score'; rl.runtimelegend = 'execution time'; rl.iterlegend = 'iterations';
rl.optlegend = 'optimal score';
rl.perfprefix = ''; 
labels = rl;    % no psfrag labels

figh = [];

if cfg.perf,
    if isscalar(cfg.M) || cfg.avgM, % only supports vector N, scalar M (should support the other way around too)
        if cfg.addtocrt,    figh(end+1) = gcf;
        else                figh(end+1) = figurex('box');
        end;
        lh = []; leg = {}; setfigprop(cfg);
        if cfg.avgM,
            eJ = reshape(data.eJ, length(cfg.N), length(cfg.M));
            eJ = mean(eJ, 2);
        else
            eJ = data.eJ;
        end;
        lh(end+1) = feval(cfg.plotfun, cfg.N, eJ, sty.perfstyle{:}); 
        hold on;
        leg{end+1} = [cfg.legendprefix labels.perflegend cfg.legendpostfix];
        if ~isempty(cfg.Jstar),
            lh(end+1) = feval(cfg.plotfun, [cfg.N(1) cfg.N(end)], [cfg.Jstar cfg.Jstar], sty.optstyle{:});
            leg{end+1} = labels.optlegend;
        end;
        annotatefig('perf', cfg, lh, leg, labels, commonprop);
    else        % 3-D plot
        if cfg.addtocrt,    figh(end+1) = gcf;
        else                figh(end+1) = figurex;
        end;
        feval(cfg.plot3fun, cfg.N, cfg.M, ...
            reshape(data.eJ, length(cfg.N), length(cfg.M))', sty.perf3style{:}); 
        setfigprop(cfg);
        colormap(sty.cm);
        xlabel(labels.x);
        ylabel(labels.u);
        zlabel(labels.perf);
        Nstep = ceil((cfg.N(end) - cfg.N(1)) / 8);
        Nticks = cfg.N(1):Nstep:cfg.N(end);
        if Nticks(end) ~= cfg.N(end), Nticks = [Nticks cfg.N(end)]; end;
        set(gca, 'XTick', Nticks);
        set(gca, 'YTick', cfg.M);
    end;
        
    varargout = {data.eJ};
end;

if cfg.runtime,
    if cfg.addtocrt,    figh(end+1) = gcf;
    else                figh(end+1) = figurex('box');
    end;
    lh = []; leg = {}; setfigprop(cfg);
    lh(end+1) = feval(cfg.plotfun, cfg.N, data.runtime, sty.runtimestyle{:}); 
    hold on;
    leg{end+1} = [cfg.legendprefix labels.runtimelegend cfg.legendpostfix];
    annotatefig('runtime', cfg, lh, leg, labels, commonprop);
end;

if cfg.iter,
    if cfg.addtocrt,    figh(end+1) = gcf;
    else                figh(end+1) = figurex('box');
    end;
    lh = []; leg = {}; setfigprop(cfg);
    lh(end+1) = feval(cfg.plotfun, cfg.N, data.iter, sty.iterstyle{:}); 
    hold on;
    leg{end+1} = [cfg.legendprefix labels.iterlegend cfg.legendpostfix];
    hold on;
    annotatefig('iter', cfg, lh, leg, labels, commonprop);
end;

% save last figure if indicated
if ~isempty(figh), 
    set(figh(end), 'Name', cfg.datafile, 'NumberTitle', 'off');
    saveplot(figh(end), [cfg.savedir cfg.savefig], cfg.plottarget);
end;


end     % fuzzyqi_resolutionstudy =============================================


% --------------------
% Local functions

function df = gen_dfname(cfg, N, M)     % generate datafile
% auto generate problem prefix in data files
if isempty(cfg.dfprefix),
    switch cfg.problem
        case 'dc_problem',          df = 'fzdc_';
        case 'ddi_problem',         df = 'fzdd_';
        case 'hillcar_problem',     df = 'fzhc_';
        case 'ipsetup_problem',     df = 'is_';
        case 'bike_problem',        df = 'fzbike_';
    end;
else    % use the given prefix 
    df = cfg.dfprefix;
end;
if ~isempty(cfg.dfinfix),
    % use given infix
    df = [df cfg.dfinfix '_'];
else
    % use first model parameter (reward type,usually) as a default infix
    if ~isempty(cfg.model_params) && ischar(cfg.model_params{1}),
    end;
end;
% add N and M to the file name
if cfg.dfcompatibility,
    df = [df num2str(N-1) 'x' num2str(N-1) '_fuzzy'];
else
    df = [df 'N' num2str(N) 'M' num2str(M)];
end;
end

% modification 31 aug 2009: removed customization of horizontal labels in
% each-axis mode (since it's redundant)
function annotatefig(what, cfg, lh, leg, labels, commonprop)

xlabel(labels.x); ylabel(labels.(what));
if cfg.eachaxis,
else
    Nstep = ceil((cfg.N(end) - cfg.N(1)) / 8);
    Nticks = cfg.N(1):Nstep:cfg.N(end);
    if Nticks(end) ~= cfg.N(end), Nticks = [Nticks cfg.N(end)]; end;
    set(gca, 'XTick', Nticks);
    tl = {};  for i = Nticks, tl{end+1} = [num2str(i) '^2']; end;
    set(gca, 'XTickLabel', tl); set(twxticklabel, commonprop{:});    
    % 2009-10-28: removed LaTeX labels (commented code below)
%     set(gca, 'XGrid', 'on');
%     for i = Nticks, tl{end+1} = ['$' num2str(i) '^2$']; end;
%     set(gca, 'XTickLabel', tl); set(twxticklabel, commonprop{:});
%     set(gca, 'XGrid', 'on');
end;
% APPEND any legend instead of replacing
if ~isempty(leg),
    [dummy1 dummy2 oldobjecth legold] = legend; clear dummy1 dummy2;
    if ~isempty(legold), lhh = legend([oldobjecth(:); lh(:)], legold{:}, leg{:}); 
    else lhh = legend(lh(:), leg{:}); end;
    set(lhh, 'Location', 'Best');
end;
setfigprop(cfg);
end     % annotatefig

% xlabel(labels.x); ylabel(labels.(what));
% N = cfg.N;
% Nstep = ceil((cfg.N(end) - cfg.N(1)) / 8);
% Nticks = cfg.N(1):Nstep:cfg.N(end);
% if Nticks(end) ~= cfg.N(end), Nticks = [Nticks cfg.N(end)]; end;
% set(gca, 'XTick', Nticks);
% tl = {}; 
% for i = Nticks; 
%     if cfg.eachaxis,    tl{end+1} = num2str(i); 
%     else                tl{end+1} = ['$' num2str(i) '^2$']; 
%     end;
% end;
% if cfg.eachaxis,    % latex not necessary
%     set(gca, 'XTickLabel', tl);
% else
%     set(gca, 'XTickLabel', tl); set(twxticklabel, commonprop{:});
% end;
% set(gca, 'XGrid', 'on');
% % APPEND any legend instead of replacing
% if ~isempty(leg),
%     [dummy1 dummy2 oldobjecth legold] = legend; clear dummy1 dummy2;
%     if ~isempty(legold), lhh = legend([oldobjecth(:); lh(:)], legold{:}, leg{:}); 
%     else lhh = legend(lh(:), leg{:}); end;
%     set(lhh, 'Location', 'Best');
% end;
% setfigprop(cfg);


