function [J, R] = eval_policy(cfg, model)
% Empirical policy evaluation on a given grid of initial states.
%   [J, R] = EVAL_POLICY(CFG, MODEL)
% Empirical policy evaluation on a given grid of initial states.
% Uses finite-length approximations of infinite returns.
% Episodic tasks are supported.
%
% Parameters:
%   CFG     - configuration parameters. See below for fields
%   MODEL   - problem model
%
% Returns:
%   J       - performance index, average of empirical returns over the
%   initial states grid
%   R       - a matrix of individual returns for every initial state in X0

% Note provides no support for dynamic controllers
% 2009-02-25: Updated to support new-style policy with t argument

% Parameters and defaults
CFG.eps = .01;              % precision for return evaluation
CFG.gamma = 0.98;           % discount
CFG.policy = @lsf;     
                            % policy function
CFG.policyargs = {};        % extra policy arguments (besides the state x and time t)
CFG.X0 = {0, 0};            % set of initial states to take into account when computing the cost
CFG.maxr = 100;             % maximum possible reward
CFG.displaypercent = 5;     % feedback

if ischar(cfg), cfg = str2cfg(cfg, fieldnames(CFG)); end;
cfg = checkparams(cfg, CFG);

N0 = [];
for i = 1:length(cfg.X0),
    N0(end+1) = length(cfg.X0{i});
end;
X0 = flat(cfg.X0);  % initial states on columns, a wide matrix
n0 =size(X0, 2);    % number of states to start from

% length of a trajectory
K = ceil(log(cfg.eps * (1-cfg.gamma) / cfg.maxr) / log(cfg.gamma));

% precompute discounting vector, pre-initialize reward vector
disc = cumprod([1 cfg.gamma+zeros(1, K)]);
r = zeros(K, 1);

R = zeros(n0, 1);
prog = cfg.displaypercent / 100;
for i = 1:n0,
    x = X0(:, i);
    for k = 1:K,
        [x r(k) terminal] = feval(model.fun, model, x, ...
            feval(cfg.policy, x, (k-1)*Ts, cfg.policyargs{:}));
        if terminal, break; end;
    end;
    % return of x0 is inner product of discounting and rewards
    R(i) = disc(1:k) * r(1:k);
    if i/n0 >= prog,
        disp([num2str(prog * 100) '% done']);
        prog = prog + cfg.displaypercent / 100;
    end;
end;
% expected value is the mean
J = mean(R);
R = reshape(R, N0);

% END rbfmceval() RETURNING expected cost eJ, cost vector J
% ===========================