% Control the DC motor - ss framework
% (c) Robert Babuska, June 2007

printflag = 0;
rtflag = 0;

h = 0.005;                          % sampling period
umax = 10;                          % maximum allowed control 

% sensors to read [status, time, angle, velocity, current, voltage, dangle/dt]
sens = [3];

%% Design parameters
w = 60;                             % desired CL natural frequency
z = 1;                              % desired CL relative damping
refmod = tf(w^2,[1 2*z*w w^2]);     % reference model

%% Model parameters
%tau = tau_est;
%km = km_est;
tau = 0.108;
km = 18.36;

DC_params;

A = [0 1 0; 0 -b/J K/J; 0 -K/L -R/L]; 
B = [0; 0; 1/L];
C = eye(3);
C = [1 0 0; 0 0 1];
C = [1 0 0];
D = 0;

% 2nd order
A = [0 1; 0 -b/J-(K^2)/(R*J)]; 
B = [0; K/(R*J)];
C = eye(2);
%C = [1 0 0; 0 0 1];
C = [1 0];
D = 0;

sys = ss(A,B,C,D);

%% Compute controller gains
Kp = tau*w^2/km;
Kd = (2*z*w*tau - 1)/km;

%% Observer design parameters
w = 150;                                        % natural frequency
z = 1;                                          % relative damping
a = 300;                                       % faster pole
refmodel = tf(w^2,[1 2*z*w w*w])*tf(a,[1 a]);   % reference model
refmodel = tf(w^2,[1 2*z*w w*w]);               % reference model
Po = pole(c2d(refmodel,h));                     % desired closed-loop poles

%% Design observer
%[Aa,Bb,Cc,Dd] = linmod('motor_model_ss');
%sys = ss(Aa,Bb,Cc,Dd);                         % create LTI object
%sys = sys(:,1);                                % ignore disturbance input
sysd = c2d(sys,h);                              % discretize
Ad = sysd.A; Bd = sysd.B; C = sysd.C;           % get system matrices
n = size(Ad,1);                                 % order of system

Ko = place(Ad',C',Po)';                         % observer gain 

%% initialize DCMOTOR - USB init
mops('CloseAll');                % close port
H = mops('Open', 'mops1');    % open port
H.WatchdogTimeout = 10;             % watchdog timeout
mops('SetParams', H);            % set the parameter
mops('Write', H, 1, 1, 0, 0);    % dummy write to sync interface board
mops('Write', H, 1, 1, 0, 0);    % reset position
mops('Write', H, 0, 1, 0, 0);    % end reset
for i = 1 : 100,                    % read a couple of times
    mops('Read',H);              % dummy read sensor data
end;

%% initialize RT
if rtflag,
    rtclear all; rtunload;
    rtload('mf624');                % load driver
    rtdef(1,'in',h,9);              % input timer
    rtdef(2,'out',h,1);             % output timer
    rtdef(3,'out',h,18);            % enable timer
    rtstart all
    rtwr(3,'y',1);
end;

%% Define reference
load ramp r 
%r = U(1:200);
t = h*(0:length(r)-1)';             % time vector

%% History variables
D1 = zeros(7,1); yy1 = 0; r1 = 0; yyf = 0;      % initialize
y = zeros(length(t),length(sens));              % initialize output history
x = zeros(length(t),n);                         % initialize state history
u = zeros(length(t),1);                         % initialize input history

%% Control loop
tic;                                            % reset Matlab's tic-toc timer
for i = 1 : length(r),                          % run 
    if rtflag,
        yy = (2*pi/2000)*rtrd(1,'y');           % read sensor data
    else
        D = mops('Read',H);                  % read sensor data
        D = [D(1:6); (D(3)-D1(3))/D(2)];        % add speed speed as dangle/dt
        D1 = D;                                 % store current measurements
        y(i,:) = D(sens);                       % select desired sensors
        Y(i,:) = D([3:5]);                      % select desired sensors
    end;
    
    % observer - correction
    x(i,:) = (x(i,:)' + Ko*(y(i,:)' - C*x(i,:)'))';

    % controller
    yy(i,:) = y(i,1);
    yy(i,:) = x(i,1);
    e(i,:) = (r(i)-yy(i));
	dy(i) = -D(4);                              % velocity provided by dcmotor
	dy(i) = -x(i,2);                            % velocity from observer
    dr = r(i) - r1;
    u(i) = Kp*e(i) + Kd*dy(i) + (Kd+18)*dr;          % control law
    u(i) = min(umax,max(-umax,u(i)));           % saturate control action
    r1 = r(i);

    % observer - prediction
    x(i+1,:) = Ad*x(i,:)' + Bd*u(i,:)';
    
    if rtflag,
        rtwr(2,'y',u(i)/10);                    % send control input to process
    else
        mops('Write',H,0,1,u(i),0);          % send control input to process
    end;

    while toc < h, end;                         % synchronize with real time
    tic;                                        % reset Matlab's tic-toc timer
end;
if rtflag,
    rtwr(3,'y',0);                              % disable
end;

% reset actuators, close the port
mops('Write', H, 0, 1, 0, 0.0);
mops('CloseAll');

x(end,:) = [];

%% Compare with model
yref = lsim(refmod,r,t);        % simulate the reference model
figure(1); clf;
subplot(211)
stairs(t,r,'k-');               % compare model with real time
hold on
stairs(t,yref,'r-');           % compare model with real time
stairs(t,yy,'b-');             % compare model with real time
hold off
%set(gca,'ylim',[-.6 .6]);
title('Closed-loop response');
xlabel('Time [s]'); 
ylabel('Reference and actual angle [rad]')

subplot(212)
stairs(t,u,'-');                      % plot u
title('Closed-loop response');
xlabel('Time [s]'); 
ylabel('Control input [-]')
set(gca,'ylim',umax*[-1 1]);
if printflag,
    print('-depsc',['angle' '_' ctype]);
end;    


if exist('v','var'),
figure(2); clf;
subplot(211)
stairs(t,v,'b-');               % compare model with real time
%set(gca,'ylim',[-.6 .6]);
title('Closed-loop response');
xlabel('Time [s]'); 
ylabel('Velocity [rad/s]')

subplot(212)
stairs(t,u,'-');                      % plot u
title('Closed-loop response');
xlabel('Time [s]'); 
ylabel('Control input [-]')
set(gca,'ylim',umax*[-1 1]);
if printflag,
    print('-depsc',['velocity' '_' ctype]);
end;    
end;

if exist('V','var'),
figure(3); clf;
subplot(211)
stairs(t,I,'b-');               % compare model with real time
%set(gca,'ylim',[-.6 .6]);
title('Closed-loop response');
xlabel('Time [s]'); 
ylabel('Current [mA]')

subplot(212)
hold on
stairs(t,u,'b-');                      % plot u
stairs(t,V,'r-');                      % plot V
hold off
title('Closed-loop response');
xlabel('Time [s]'); 
ylabel('Control input [-] and voltage [V]')
set(gca,'ylim',umax*[-1 1]);
if printflag,
    print('-depsc',['current' '_' ctype]);
end;    
end;
