function rarm_lqr(cfg)
% Run LQR reward fuzzy Q-iteration for robotic arm, and plot results
% This function was written for the Fuzzy Q-iteration jnl paper

% The model should be the same as in the FUZZ-IEEE-07 paper, just the reward function is
% changed
% Perhaps should run with ode45 integration, as well, to make sure results are OK. (currently
% ran with ode4/5 steps per sample time)

% Note this experiment uses no "rollover" MFs (but the state vars DO roll over)

if nargin < 1, cfg = struct; end;

% The config below should not be touched, since it's used by the thesis
CFG.savedir = 'D:\Work\tex\papers\fzqi\img\';
CFG.plottarget = 'screen';
CFG.fname = 'rarm_fz_lqr';
CFG.loadapprox = [];
CFG.serial = 0;
CFG.overwrite = 0;              % overwrite data file even if exists
CFG.infix = '';                 % infix for saved figures 
% number of fuzzy sets, discrete actions, and their spacing
CFG.spacing = 'log';
CFG.N = [13 7 13 7];
CFG.M = [5 5];
CFG.interph = 1;
% replay, trajectory configuration 
% NOTE this mechanism has been superseded, use trajopt instead
CFG.unwrap = 0;                 % whether to unwrap trajectories
CFG.replayopt.x0 = [pi, 0, -pi, 0];       % initial state for replay
CFG.trajopt = struct;
CFG.trajfigsize = [650 700];    % made tunable for thesis chapter plot
CFG.hfigsize = [];              % made tunable for book chapter
CFG.modelcfg = struct;

CFG.colorbar = 1;
CFG.subplotorient = 'v';
% install user-defined params
cfg = parseconfig(cfg, CFG);

% test for vertical orientation
% cfg.N = [7 5 7 5];            % discretization spacing
% cfg.M = [3 3];
% cfg.fname = 'rarm_testvert';
% cfg.modelcfg.orientation = 'vert';
% cfg.modelcfg.maxtau = [5 1];  % this seems to be sufficient for pushing up
% cfg.overwrite = 1;

% compute default grids (use the modelcfg above to make sure they are correct)
model = rarm_problem('model', cfg.modelcfg);  % required to get X and U bounds
switch cfg.spacing,
    case 'log';
        gridfun = @symloggrid;
    case 'equid',
        gridfun = @symequidgrid;
    otherwise,
        error('unknown spacing');
end;        
GRIDCFG.xgrids = {...
        gridfun(cfg.N(1), pi), gridfun(cfg.N(2), model.maxomega(1)), ...
        gridfun(cfg.N(3), pi), gridfun(cfg.N(4), model.maxomega(2)) };
GRIDCFG.ugrids = {...
        gridfun(cfg.M(1), model.maxtau(1)), ...
        gridfun(cfg.M(2), model.maxtau(2)) };
% install any explicitly user-defined grids
cfg = parseconfig (cfg, GRIDCFG);

if ~exist([cfg.fname '.mat'], 'file') || cfg.overwrite,
    runcfg = struct;
    runcfg.run = 1;
    runcfg.problem = 'rarm_problem';
    runcfg.model_params = {cfg.modelcfg};
    runcfg.gamma = 0.98;
    runcfg.xgrids = cfg.xgrids;
    runcfg.ugrids = cfg.ugrids;
    runcfg.eps = 1e-5;
    runcfg.maxiter = 1000;
    runcfg.loadapprox = cfg.loadapprox;
    runcfg.serial = cfg.serial;
    runcfg.datafile = cfg.fname;
    runcfg.verb = 10;
    runcfg.initstepdisp = 0.1/3;

    fuzzyqi(runcfg);
end;

rcfg = struct;
rcfg.replay = 1;
rcfg.datafile = cfg.fname;
rcfg.interph = cfg.interph;
rcfg.silent = 1;
rcfg.tend = 5;
% any custom fields
rcfg = copyfields(cfg.replayopt, rcfg, fieldnames(cfg.replayopt));
hist = fuzzyqi(rcfg);
if cfg.unwrap,
    hist.x(1, :) = unwrapangle(hist.x(1, :), 0, 0.5);
    hist.x(3, :) = unwrapangle(hist.x(3, :), 0, 0.5);
end;

rcfg = struct;
rcfg.plottarget = cfg.plottarget;
rcfg.savedir = cfg.savedir;
if isempty(cfg.infix), 
    rcfg.savefig = 'rarm_traj';
else
    rcfg.savefig = ['rarm_' cfg.infix '_traj'];
end;
rcfg.hist = hist;
rcfg.trajectory = 1;
rcfg.figsize = cfg.trajfigsize;
rcfg = copyfields(cfg.trajopt, rcfg, fieldnames(cfg.trajopt));
rarm_plot(rcfg);

pcfg = struct;
pcfg.fuzzyh = 1;
pcfg.datafile = cfg.fname;
pcfg.fzstyle = 'qi';    % use the new, approxrl standard style for variables in plot fun
pcfg.basis = 1;
pcfg.gridres = 120;
pcfg.markersize = 6;
pcfg.interph = cfg.interph;
pcfg.plottarget = cfg.plottarget;
pcfg.savedir = cfg.savedir;
if isempty(cfg.infix), 
    pcfg.savefig = 'rarm_h';
else
    pcfg.savefig = ['rarm_' cfg.infix '_h'];
end;
pcfg.colorbar = cfg.colorbar;
pcfg.subplotorient = cfg.subplotorient;
if isempty(cfg.hfigsize),
    if cfg.subplotorient == 'v',    pcfg.figsize = [450 800];
    else                            pcfg.figsize = [700 300];
    end;    
else
    pcfg.figsize = cfg.hfigsize;
end;
rarm_plot(pcfg);
